from future import standard_library
standard_library.install_aliases()
from builtins import str  # @IgnorePep8
import os.path  # @IgnorePep8
# from nipype import config
# config.enable_debug_mode()
from arcana.data.file_format import text_format  # @IgnorePep8
from arcana.study.base import Study, StudyMetaClass  # @IgnorePep8
from arcana.utils.testing import (  # @IgnorePep8
    BaseTestCase, BaseMultiSubjectTestCase, TestMath)  # @IgnorePep8
from arcana.exceptions import (  # @IgnorePep8
    ArcanaCantPickleStudyError, ArcanaUsageError)  # @IgnorePep8
from arcana.study.multi import (  # @IgnorePep8
    MultiStudy, MultiStudyMetaClass, SubStudySpec)
from nipype.interfaces.base import (  # @IgnorePep8
    BaseInterface, File, TraitedSpec, traits, isdefined)
from arcana.study.parameter import ParamSpec  # @IgnorePep8
from arcana.data.file_format import FileFormat, IdentityConverter  # @IgnorePep8
from nipype.interfaces.utility import IdentityInterface  # @IgnorePep8
from arcana.exceptions import ArcanaNoConverterError  # @IgnorePep8
from arcana.repository import Tree  # @IgnorePep8
from arcana.data import (  # @IgnorePep8
    Fileset, FieldSpec, InputFilesetSpec, InputFilesets, FilesetSpec)
from future.utils import PY2  # @IgnorePep8
from future.utils import with_metaclass  # @IgnorePep8
import logging  # @IgnorePep8
if PY2:
    import pickle as pkl  # @UnusedImport
else:
    import pickle as pkl  # @Reimport


class ExampleStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('one', text_format),
        InputFilesetSpec('ten', text_format),
        FilesetSpec('derived1_1', text_format, 'pipeline1'),
        FilesetSpec('derived1_2', text_format, 'pipeline1'),
        FilesetSpec('derived2', text_format, 'pipeline2'),
        FilesetSpec('derived3', text_format, 'pipeline3'),
        FilesetSpec('derived4', text_format, 'pipeline4'),
        FieldSpec('derived5a', str, 'pipeline5',
                  pipeline_args={'arg': 'a'}),
        FieldSpec('derived5b', str, 'pipeline5',
                  pipeline_args={'arg': 'b'}),
        FilesetSpec('subject_summary', text_format,
                    'subject_summary_pipeline',
                    frequency='per_subject'),
        FilesetSpec('visit_summary', text_format,
                    'visit_summary_pipeline',
                    frequency='per_visit'),
        FilesetSpec('study_summary', text_format,
                    'study_summary_pipeline',
                    frequency='per_study'),
        FilesetSpec('subject_ids', text_format,
                    'subject_ids_access_pipeline',
                    frequency='per_visit'),
        FilesetSpec('visit_ids', text_format,
                    'visit_ids_access_pipeline',
                    frequency='per_subject')]

    add_param_specs = [
        ParamSpec('pipeline_parameter', False)]

    def pipeline1(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline1',
            desc="A dummy pipeline used to test 'run_pipeline' method",
            citations=[],
            name_maps=name_maps)
        if not self.parameter('pipeline_parameter'):
            raise Exception("Pipeline parameter was not accessible")
        indent = pipeline.add("ident1", IdentityInterface(['file']))
        indent2 = pipeline.add("ident2", IdentityInterface(['file']))
        # Connect inputs
        pipeline.connect_input('one', indent, 'file')
        pipeline.connect_input('one', indent2, 'file')
        # Connect outputs
        pipeline.connect_output('derived1_1', indent, 'file')
        pipeline.connect_output('derived1_2', indent2, 'file')
        return pipeline

    def pipeline2(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline2',
            desc="A dummy pipeline used to test 'run_pipeline' method",
            citations=[],
            name_maps=name_maps)
        if not self.parameter('pipeline_parameter'):
            raise Exception("Pipeline parameter was not cascaded down to "
                            "pipeline2")
        math = pipeline.add("math", TestMath())
        math.inputs.op = 'add'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('one', math, 'x')
        pipeline.connect_input('derived1_1', math, 'y')
        # Connect outputs
        pipeline.connect_output('derived2', math, 'z')
        return pipeline

    def pipeline3(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline3',
            desc="A dummy pipeline used to test 'run_pipeline' method",
            citations=[],
            name_maps=name_maps)
        indent = pipeline.add('ident', IdentityInterface(['file']))
        # Connect inputs
        pipeline.connect_input('derived2', indent, 'file')
        # Connect outputs
        pipeline.connect_output('derived3', indent, 'file')
        return pipeline

    def pipeline4(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline4',
            desc="A dummy pipeline used to test 'run_pipeline' method",
            citations=[],
            name_maps=name_maps)
        math = pipeline.add("mrcat", TestMath())
        math.inputs.op = 'mul'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('derived1_2', math, 'x')
        pipeline.connect_input('derived3', math, 'y')
        # Connect outputs
        pipeline.connect_output('derived4', math, 'z')
        return pipeline

    def pipeline5(self, arg, **name_maps):

        pipeline = self.new_pipeline(
            name='pipeline5{}'.format(arg),
            desc="A dummy pipeline used to test constructor arguments",
            citations=[],
            name_maps=name_maps)

        pipeline.add(
            "ident",
            IdentityInterface(
                fields=['value', 'dummy'],
                value=arg),
            inputs={
                'dummy': ('one', text_format)},
            outputs={
                'derived5{}'.format(arg): ('value', str)})

        return pipeline

    def visit_ids_access_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='visit_ids_access',
            desc=(
                "A dummy pipeline used to test access to 'session' IDs"),
            citations=[],
            name_maps=name_maps)
        visits_to_file = pipeline.add(
            'visits_to_file', IteratorToFile(), joinsource=self.VISIT_ID,
            joinfield='ids')
        pipeline.connect_input(self.VISIT_ID, visits_to_file, 'ids')
        pipeline.connect_input(self.SUBJECT_ID, visits_to_file, 'fixed_id')
        pipeline.connect_output('visit_ids', visits_to_file, 'out_file')
        return pipeline

    def subject_ids_access_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='subject_ids_access',
            desc=(
                "A dummy pipeline used to test access to 'subject' IDs"),
            citations=[],
            name_maps=name_maps)
        subjects_to_file = pipeline.add(
            'subjects_to_file', IteratorToFile(), joinfield='ids',
            joinsource=self.SUBJECT_ID)
        pipeline.connect_input(self.SUBJECT_ID, subjects_to_file, 'ids')
        pipeline.connect_input(self.VISIT_ID, subjects_to_file, 'fixed_id')
        pipeline.connect_output('subject_ids', subjects_to_file, 'out_file')
        return pipeline

    def subject_summary_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name="subject_summary",
            desc=("Test of project summary variables"),
            citations=[],
            name_maps=name_maps)
        math = pipeline.add(
            'math', TestMath(), joinfield='x', joinsource=self.VISIT_ID)
        math.inputs.op = 'add'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('one', math, 'x')
        # Connect outputs
        pipeline.connect_output('subject_summary', math, 'z')
        return pipeline

    def visit_summary_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name="visit_summary",
            desc=("Test of project summary variables"),
            citations=[],
            name_maps=name_maps)
        math = pipeline.add('math', TestMath(), joinfield='x',
                            joinsource=self.SUBJECT_ID)
        math.inputs.op = 'add'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('one', math, 'x')
        # Connect outputs
        pipeline.connect_output('visit_summary', math, 'z')
        return pipeline

    def study_summary_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name="study_summary",
            desc=("Test of project summary variables"),
            citations=[],
            name_maps=name_maps)
        math1 = pipeline.add(
            'math1', TestMath(), joinfield='x', joinsource=self.VISIT_ID)
        math2 = pipeline.add(
            'math2', TestMath(), joinfield='x', joinsource=self.SUBJECT_ID)
        math1.inputs.op = 'add'
        math2.inputs.op = 'add'
        math1.inputs.as_file = True
        math2.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('one', math1, 'x')
        pipeline.connect(math1, 'z', math2, 'x')
        # Connect outputs
        pipeline.connect_output('study_summary', math2, 'z')
        return pipeline


class IteratorToFileInputSpec(TraitedSpec):
    ids = traits.List(traits.Str(), desc="ID of the iterable")
    fixed_id = traits.Str(desc="The other ID that will remain fixed")
    out_file = File(genfile=True, desc="The name of the generated file")


class IteratorToFileOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Output file containing iterables")


class IteratorToFile(BaseInterface):

    input_spec = IteratorToFileInputSpec
    output_spec = IteratorToFileOutputSpec

    def _run_interface(self, runtime):
        with open(self._gen_filename('out_file'), 'w') as f:
            f.write('\n'.join(str(i) for i in self.inputs.ids))
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            if isdefined(self.inputs.out_file):
                fname = self.inputs.out_file
            else:
                fname = os.path.join(os.getcwd(), 'out.txt')
        else:
            assert False
        return fname


class TestStudy(BaseMultiSubjectTestCase):

    SUBJECT_IDS = ['SUBJECTID1', 'SUBJECTID2', 'SUBJECTID3']
    VISIT_IDS = ['VISITID1', 'VISITID2']
    DATASET_CONTENTS = {'one_input': '1', 'ten_input': '10'}

    @property
    def input_tree(self):
        filesets = []
        for subj_id in self.SUBJECT_IDS:
            for visit_id in self.VISIT_IDS:
                filesets.append(
                    Fileset('one_input', text_format,
                            subject_id=subj_id, visit_id=visit_id))
                filesets.append(
                    Fileset('ten_input', text_format,
                            subject_id=subj_id,
                            visit_id=visit_id))
        return Tree.construct(self.repository, filesets=filesets)

    def make_study(self):
        return self.create_study(
            ExampleStudy, 'dummy', inputs=[
                InputFilesets('one', 'one_input', text_format),
                InputFilesets('ten', 'ten_input', text_format)],
            parameters={'pipeline_parameter': True})

    def test_run_pipeline_with_prereqs(self):
        study = self.make_study()
        self.assertContentsEqual(study.data('derived4'),
                                 [2.0, 2.0, 2.0, 2.0, 2.0, 2.0])

    def test_pipeline_args(self):
        study = self.make_study()
        a = next(iter(study.data('derived5a'))).value
        b = next(iter(study.data('derived5b'))).value
        self.assertEqual(a, 'a')
        self.assertEqual(b, 'b')

    def test_subject_summary(self):
        study = self.make_study()
        summaries = study.data('subject_summary')
        ref = float(len(self.VISIT_IDS))
        for fileset in summaries:
            self.assertContentsEqual(fileset, ref,
                                     str(fileset.visit_id))

    def test_visit_summary(self):
        study = self.make_study()
        summaries = study.data('visit_summary')
        ref = float(len(self.SUBJECT_IDS))
        for fileset in summaries:
            self.assertContentsEqual(fileset, ref,
                                     str(fileset.visit_id))

    def test_study_summary(self):
        study = self.make_study()
        ref = float(len(self.SUBJECT_IDS) * len(self.VISIT_IDS))
        self.assertContentsEqual(study.data('study_summary'), ref)

    def test_subject_ids_access(self):
        study = self.make_study()
        study.data('subject_ids')
        for visit_id in self.VISIT_IDS:
            subject_ids_path = self.output_file_path(
                'subject_ids.txt', study.name,
                visit=visit_id, frequency='per_visit')
            with open(subject_ids_path) as f:
                ids = f.read().split('\n')
            self.assertEqual(sorted(ids), sorted(self.SUBJECT_IDS))

    def test_visit_ids_access(self):
        study = self.make_study()
        study.data('visit_ids')
        for subject_id in self.SUBJECT_IDS:
            visit_ids_path = self.output_file_path(
                'visit_ids.txt', study.name,
                subject=subject_id, frequency='per_subject')
            with open(visit_ids_path) as f:
                ids = f.read().split('\n')
            self.assertEqual(sorted(ids), sorted(self.VISIT_IDS))


class ExistingPrereqStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('one', text_format),
        FilesetSpec('ten', text_format, 'ten_pipeline'),
        FilesetSpec('hundred', text_format, 'hundred_pipeline'),
        FilesetSpec('thousand', text_format, 'thousand_pipeline')]

    def pipeline_factory(self, incr, input, output, name_maps):  # @ReservedAssignment @IgnorePep8
        pipeline = self.new_pipeline(
            name=output + '_pipeline',
            desc="A dummy pipeline used to test 'partial-complete' method",
            citations=[], name_maps=name_maps)
        # Nodes
        math = pipeline.add("math", TestMath())
        math.inputs.y = incr
        math.inputs.op = 'add'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input(input, math, 'x')
        # Connect outputs
        pipeline.connect_output(output, math, 'z')
        return pipeline

    def ten_pipeline(self, **name_maps):  # @UnusedVariable
        return self.pipeline_factory(10, 'one', 'ten', name_maps=name_maps)

    def hundred_pipeline(self, **name_maps):  # @UnusedVariable
        return self.pipeline_factory(100, 'ten', 'hundred',
                                     name_maps=name_maps)

    def thousand_pipeline(self, **name_maps):  # @UnusedVariable
        return self.pipeline_factory(1000, 'hundred', 'thousand',
                                     name_maps=name_maps)


class TestExistingPrereqs(BaseMultiSubjectTestCase):
    """
    This unittest tests out that partially previously calculated prereqs
    are detected and not rerun unless reprocess==True.
    """

    PROJECT_STRUCTURE = {
        'subject1': {
            'visit1': ['one', 'ten', 'hundred'],
            'visit2': ['one', 'ten'],
            'visit3': ['one', 'ten', 'hundred', 'thousand']},
        'subject2': {
            'visit1': ['one'],
            'visit2': ['one', 'ten'],
            'visit3': ['one', 'ten', 'hundred', 'thousand']}}

    DATASET_CONTENTS = {'one': 1.0, 'ten': 10.0, 'hundred': 100.0,
                        'thousand': 1000.0}

    STUDY_NAME = 'study'

    @property
    def input_tree(self):
        filesets = []
        for subj_id, visit_ids in list(self.PROJECT_STRUCTURE.items()):
            for visit_id, fileset_names in list(visit_ids.items()):
                # Create filesets
                for name in fileset_names:
                    from_study = self.STUDY_NAME if name != 'one' else None
                    filesets.append(
                        Fileset(name, text_format, subject_id=subj_id,
                                visit_id=visit_id, from_study=from_study))
        return Tree.construct(self.repository, filesets=filesets)

    def add_sessions(self):
        BaseMultiSubjectTestCase.add_sessions(self)
        # Create a study object, in order to generate appropriate provenance
        # for the existing "derived" data
        derived_filesets = [f for f in self.DATASET_CONTENTS
                            if f != 'one']
        study = self.create_study(
            ExistingPrereqStudy, self.STUDY_NAME,
            repository=self.local_repository,
            inputs=[InputFilesets('one', 'one', text_format)])
        # Get all pipelines in the study
        pipelines = {n: getattr(study, '{}_pipeline'.format(n))()
                     for n in derived_filesets}
        for node in study.tree:
            for fileset in node.filesets:
                if fileset.basename != 'one' and fileset.exists:
                    # Generate expected provenance record for each pipeline
                    # and save in the local repository
                    pipelines[fileset.name].cap()
                    record = pipelines[fileset.name].expected_record(node)
                    self.local_repository.put_record(record)
        study.clear_caches()  # Reset repository trees

    def test_per_session_prereqs(self):
        # Generate all data for 'thousand' spec
        study = self.create_study(
            ExistingPrereqStudy, self.STUDY_NAME,
            inputs=[InputFilesets('one', 'one', text_format)])
        study.data('thousand')
        targets = {
            'subject1': {
                'visit1': 1100.0,
                'visit2': 1110.0,
                'visit3': 1000.0},
            'subject2': {
                'visit1': 1111.0,
                'visit2': 1110.0,
                'visit3': 1000.0}}
        tree = self.repository.tree()
        for subj_id, visits in self.PROJECT_STRUCTURE.items():
            for visit_id in visits:
                session = tree.subject(subj_id).session(visit_id)
                fileset = session.fileset('thousand',
                                          from_study=self.STUDY_NAME)
                fileset.format = text_format
                self.assertContentsEqual(
                    fileset, targets[subj_id][visit_id],
                    "{}:{}".format(subj_id, visit_id))


test1_format = FileFormat('test1', extension='.t1')
test2_format = FileFormat('test2', extension='.t2')
test3_format = FileFormat('test3', extension='.t3')

test2_format.set_converter(test1_format, IdentityConverter)


class TestInputValidationStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('a', (test1_format, test2_format)),
        InputFilesetSpec('b', test3_format),
        FilesetSpec('c', test2_format, 'identity_pipeline'),
        FilesetSpec('d', test3_format, 'identity_pipeline')]

    def identity_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline',
            desc="A dummy pipeline used to test study input validation",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a', 'b']))
        pipeline.connect_input('a', identity, 'a')
        pipeline.connect_input('b', identity, 'b')
        pipeline.connect_output('c', identity, 'a')
        pipeline.connect_output('d', identity, 'b')


class TestInputValidation(BaseTestCase):

    def setUp(self):
        self.reset_dirs()
        os.makedirs(self.session_dir)
        for spec_name, fformat in (('a', test1_format), ('b', test3_format),
                                   ('c', test1_format), ('d', test3_format)):
            with open(os.path.join(self.session_dir, spec_name) +
                      fformat.ext, 'w') as f:
                f.write(spec_name)

    def test_input_validation(self):
        self.create_study(
            TestInputValidationStudy,
            'test_input_validation',
            inputs=[
                InputFilesets('a', 'a', test1_format),
                InputFilesets('b', 'b', test3_format),
                InputFilesets('c', 'a', test1_format),
                InputFilesets('d', 'd', test3_format)])


class TestInputValidationFail(BaseTestCase):

    def setUp(self):
        self.reset_dirs()
        os.makedirs(self.session_dir)
        for spec in TestInputValidationStudy.data_specs():
            with open(os.path.join(self.session_dir, spec.name) +
                      test3_format.ext, 'w') as f:
                f.write(spec.name)

    def test_input_validation_fail(self):
        self.assertRaises(
            ArcanaUsageError,
            self.create_study,
            TestInputValidationStudy,
            'test_validation_fail',
            inputs=[
                InputFilesets('a', 'a', test3_format),
                InputFilesets('b', 'b', test3_format)])


class TestInputNoConverter(BaseTestCase):

    def setUp(self):
        self.reset_dirs()
        os.makedirs(self.session_dir)
        for spec in TestInputValidationStudy.data_specs():
            if spec.name == 'a':
                ext = test1_format.ext
            else:
                ext = test3_format.ext
            with open(os.path.join(self.session_dir, spec.name) + ext,
                      'w') as f:
                f.write(spec.name)

    def test_input_validation_fail(self):
        self.assertRaises(
            ArcanaNoConverterError,
            self.create_study,
            TestInputValidationStudy,
            'test_validation_fail',
            inputs=[
                InputFilesets('a', 'a', test1_format),
                InputFilesets('b', 'b', test3_format),
                InputFilesets('c', 'c', test3_format),
                InputFilesets('d', 'd', test3_format)])


class AlwaysRaisedError(Exception):
    pass


class ErrorInterfaceInputSpec(TraitedSpec):

    in_file = File('a file')


class ErrorInterfaceOutputSpec(TraitedSpec):

    out_file = File('a file')


class ErrorInterface(BaseInterface):

    input_spec = ErrorInterfaceInputSpec
    output_spec = ErrorInterfaceOutputSpec

    def _run_interface(self, runtime):
        raise AlwaysRaisedError


class BasicTestStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('fileset', text_format),
        FilesetSpec('out_fileset', text_format, 'a_pipeline'),
        FilesetSpec('raise_error', text_format, 'raise_error_pipeline')]

    def a_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            'a_pipeline',
            desc='a dummy pipeline',
            citations=[],
            name_maps=name_maps)

        pipeline.add(
            'ident',
            IdentityInterface(
                ['fileset']),
            inputs={
                'fileset': ('fileset', text_format)},
            outputs={
                'out_fileset': ('fileset', text_format)})

        return pipeline

    def raise_error_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            'raise_error_pipeline',
            desc='a pipeline that always throws an error',
            citations=[],
            name_maps=name_maps)

        pipeline.add(
            'error',
            ErrorInterface(),
            inputs={
                'in_file': ('fileset', text_format)},
            outputs={
                'raise_error': ('out_file', text_format)})

        return pipeline


class TestInterfaceErrorHandling(BaseTestCase):

    INPUT_FILESETS = {'fileset': 'foo'}

    def test_raised_error(self):
        study = self.create_study(
            BasicTestStudy,
            'base',
            inputs=[InputFilesets('fileset', 'fileset', text_format)])

        # Disable error logs as it should always throw an error
        logger = logging.getLogger('nipype.workflow')
        orig_level = logger.level
        logger.setLevel(50)
        self.assertRaises(
            RuntimeError,
            study.data,
            'raise_error')
        logger.setLevel(orig_level)


class TestGeneratedPickle(BaseTestCase):

    INPUT_FILESETS = {'fileset': 'foo'}

    def test_generated_cls_pickle(self):
        GeneratedClass = StudyMetaClass(
            'GeneratedClass', (BasicTestStudy,), {})
        study = self.create_study(
            GeneratedClass,
            'gen_cls',
            inputs=[InputFilesets('fileset', 'fileset', text_format)])
        pkl_path = os.path.join(self.work_dir, 'gen_cls.pkl')
        with open(pkl_path, 'wb') as f:
            pkl.dump(study, f)
        del GeneratedClass
        with open(pkl_path, 'rb') as f:
            regen = pkl.load(f)
        self.assertContentsEqual(regen.data('out_fileset'), 'foo')

    def test_multi_study_generated_cls_pickle(self):
        cls_dct = {
            'add_substudy_specs': [
                SubStudySpec('ss1', BasicTestStudy),
                SubStudySpec('ss2', BasicTestStudy)]}
        MultiGeneratedClass = MultiStudyMetaClass(
            'MultiGeneratedClass', (MultiStudy,), cls_dct)
        study = self.create_study(
            MultiGeneratedClass,
            'multi_gen_cls',
            inputs=[InputFilesets('ss1_fileset', 'fileset', text_format),
                    InputFilesets('ss2_fileset', 'fileset', text_format)])
        pkl_path = os.path.join(self.work_dir, 'multi_gen_cls.pkl')
        with open(pkl_path, 'wb') as f:
            pkl.dump(study, f)
        del MultiGeneratedClass
        with open(pkl_path, 'rb') as f:
            regen = pkl.load(f)
        self.assertContentsEqual(regen.data('ss2_out_fileset'), 'foo')

    def test_genenerated_method_pickle_fail(self):
        cls_dct = {
            'add_substudy_specs': [
                SubStudySpec('ss1', BasicTestStudy),
                SubStudySpec('ss2', BasicTestStudy)],
            'default_fileset_pipeline': MultiStudy.translate(
                'ss1', 'pipeline')}
        MultiGeneratedClass = MultiStudyMetaClass(
            'MultiGeneratedClass', (MultiStudy,), cls_dct)
        study = self.create_study(
            MultiGeneratedClass,
            'multi_gen_cls',
            inputs=[InputFilesets('ss1_fileset', 'fileset', text_format),
                    InputFilesets('ss2_fileset', 'fileset', text_format)])
        pkl_path = os.path.join(self.work_dir, 'multi_gen_cls.pkl')
        with open(pkl_path, 'w') as f:
            self.assertRaises(
                ArcanaCantPickleStudyError,
                pkl.dump,
                study,
                f)



from future import standard_library
standard_library.install_aliases()
import os  # @IgnorePep8
from nipype.pipeline import engine as pe  # @IgnorePep8
from nipype.interfaces.utility import IdentityInterface  # @IgnorePep8
from arcana.data.file_format import text_format  # @IgnorePep8
from arcana.processor import SingleProc  # @IgnorePep8
from arcana.data import (  # @IgnorePep8
    InputFilesets, FieldSpec)  # @IgnorePep8
from arcana.utils import PATH_SUFFIX  # @IgnorePep8
from future.utils import with_metaclass  # @IgnorePep8
from arcana.utils.testing import BaseTestCase  # @IgnorePep8
from arcana.data import InputFilesetSpec, FilesetSpec  # @IgnorePep8
from arcana.study import Study, StudyMetaClass  # @IgnorePep8
from arcana.repository.interfaces import RepositorySource, RepositorySink  # @IgnorePep8
from arcana.repository.basic import BasicRepo  # @IgnorePep8


class DummyStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('source1', text_format, optional=True),
        InputFilesetSpec('source2', text_format, optional=True),
        InputFilesetSpec('source3', text_format, optional=True),
        InputFilesetSpec('source4', text_format, optional=True),
        FilesetSpec('sink1', text_format, 'dummy_pipeline'),
        FilesetSpec('sink3', text_format, 'dummy_pipeline'),
        FilesetSpec('sink4', text_format, 'dummy_pipeline'),
        FilesetSpec('subject_sink', text_format, 'dummy_pipeline',
                    frequency='per_subject'),
        FilesetSpec('visit_sink', text_format, 'dummy_pipeline',
                    frequency='per_visit'),
        FilesetSpec('study_sink', text_format, 'dummy_pipeline',
                    frequency='per_study'),
        FilesetSpec('resink1', text_format, 'dummy_pipeline'),
        FilesetSpec('resink2', text_format, 'dummy_pipeline'),
        FilesetSpec('resink3', text_format, 'dummy_pipeline'),
        FieldSpec('field1', int, 'dummy_pipeline'),
        FieldSpec('field2', float, 'dummy_pipeline'),
        FieldSpec('field3', str, 'dummy_pipeline')]

    def dummy_pipeline(self, **name_maps):
        return self.new_pipeline('dummy', name_maps=name_maps)


class TestSinkAndSource(BaseTestCase):

    STUDY_NAME = 'astudy'
    SUMMARY_STUDY_NAME = 'asummary'
    INPUT_FILESETS = {'source1': '1',
                      'source2': '2',
                      'source3': '3',
                      'source4': '4'}

    def test_repository_roundtrip(self):
        study = DummyStudy(
            self.STUDY_NAME, self.repository, processor=SingleProc('a_dir'),
            inputs=[InputFilesets('source1', 'source1', text_format),
                    InputFilesets('source2', 'source2', text_format),
                    InputFilesets('source3', 'source3', text_format),
                    InputFilesets('source4', 'source4', text_format)])
        # TODO: Should test out other file formats as well.
        source_files = ('source1', 'source2', 'source3', 'source4')
        sink_files = ('sink1', 'sink3', 'sink4')
        inputnode = pe.Node(IdentityInterface(['subject_id', 'visit_id']),
                            'inputnode')
        inputnode.inputs.subject_id = self.SUBJECT
        inputnode.inputs.visit_id = self.VISIT
        source = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection
                for f in source_files),
            name='source')
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection for f in sink_files),
                dummy_pipeline),
            name='sink')
        sink.inputs.name = 'repository_sink'
        sink.inputs.desc = (
            "A test session created by repository roundtrip unittest")
        # Create workflow connecting them together
        workflow = pe.Workflow('source_sink_unit_test', base_dir=self.work_dir)
        workflow.add_nodes((source, sink))
        workflow.connect(inputnode, 'subject_id', source, 'subject_id')
        workflow.connect(inputnode, 'visit_id', source, 'visit_id')
        workflow.connect(inputnode, 'subject_id', sink, 'subject_id')
        workflow.connect(inputnode, 'visit_id', sink, 'visit_id')
        for source_name in source_files:
            if not source_name.endswith('2'):
                sink_name = source_name.replace('source', 'sink')
                workflow.connect(
                    source, source_name + PATH_SUFFIX,
                    sink, sink_name + PATH_SUFFIX)
        workflow.run()
        # Check local directory was created properly
        outputs = [
            f for f in sorted(os.listdir(
                self.get_session_dir(from_study=self.STUDY_NAME)))
            if f not in (BasicRepo.FIELDS_FNAME,
                         BasicRepo.PROV_DIR)]
        self.assertEqual(outputs, ['sink1.txt', 'sink3.txt', 'sink4.txt'])

    def test_fields_roundtrip(self):
        STUDY_NAME = 'fields_roundtrip'
        study = DummyStudy(
            STUDY_NAME, self.repository,
            processor=SingleProc('a_dir'),
            inputs=[])
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection
                 for f in ['field1', 'field2', 'field3']), dummy_pipeline),
            name='fields_sink')
        sink.inputs.field1_field = field1 = 1
        sink.inputs.field2_field = field2 = 2.0
        sink.inputs.field3_field = field3 = '3'
        sink.inputs.subject_id = self.SUBJECT
        sink.inputs.visit_id = self.VISIT
        sink.inputs.desc = "Test sink of fields"
        sink.inputs.name = 'test_sink'
        sink.run()
        source = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection
                for f in ['field1', 'field2', 'field3']),
            name='fields_source')
        source.inputs.visit_id = self.VISIT
        source.inputs.subject_id = self.SUBJECT
        source.inputs.desc = "Test source of fields"
        source.inputs.name = 'test_source'
        results = source.run()
        self.assertEqual(results.outputs.field1_field, field1)
        self.assertEqual(results.outputs.field2_field, field2)
        self.assertEqual(results.outputs.field3_field, field3)

    def test_summary(self):
        study = DummyStudy(
            self.SUMMARY_STUDY_NAME, self.repository, SingleProc('ad'),
            inputs=[InputFilesets('source1', 'source1', text_format),
                    InputFilesets('source2', 'source2', text_format),
                    InputFilesets('source3', 'source3', text_format)])
        # TODO: Should test out other file formats as well.
        source_files = ['source1', 'source2', 'source3']
        inputnode = pe.Node(
            IdentityInterface(['subject_id', 'visit_id']), 'inputnode')
        inputnode.inputs.subject_id = self.SUBJECT
        inputnode.inputs.visit_id = self.VISIT
        source = pe.Node(
            RepositorySource(study.bound_spec(f).collection
                             for f in source_files), name='source')
        # Test subject sink
        subject_sink_files = ['subject_sink']
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        subject_sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection for f in subject_sink_files),
                dummy_pipeline),
            name='subject_sink')
        subject_sink.inputs.name = 'subject_summary'
        subject_sink.inputs.desc = (
            "Tests the sinking of subject-wide filesets")
        # Test visit sink
        visit_sink_files = ['visit_sink']
        visit_sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection
                 for f in visit_sink_files),
                dummy_pipeline),
            name='visit_sink')
        visit_sink.inputs.name = 'visit_summary'
        visit_sink.inputs.desc = (
            "Tests the sinking of visit-wide filesets")
        # Test project sink
        study_sink_files = ['study_sink']
        study_sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection for f in study_sink_files),
                dummy_pipeline),
            name='study_sink')

        study_sink.inputs.name = 'project_summary'
        study_sink.inputs.desc = (
            "Tests the sinking of project-wide filesets")
        # Create workflow connecting them together
        workflow = pe.Workflow('summary_unittest', base_dir=self.work_dir)
        workflow.add_nodes((source, subject_sink, visit_sink,
                            study_sink))
        workflow.connect(inputnode, 'subject_id', source, 'subject_id')
        workflow.connect(inputnode, 'visit_id', source, 'visit_id')
        workflow.connect(inputnode, 'subject_id', subject_sink, 'subject_id')
        workflow.connect(inputnode, 'visit_id', visit_sink, 'visit_id')
        workflow.connect(
            source, 'source1' + PATH_SUFFIX,
            subject_sink, 'subject_sink' + PATH_SUFFIX)
        workflow.connect(
            source, 'source2' + PATH_SUFFIX,
            visit_sink, 'visit_sink' + PATH_SUFFIX)
        workflow.connect(
            source, 'source3' + PATH_SUFFIX,
            study_sink, 'study_sink' + PATH_SUFFIX)
        workflow.run()
        # Check local summary directories were created properly
        subject_dir = self.get_session_dir(
            frequency='per_subject',
            from_study=self.SUMMARY_STUDY_NAME)
        self.assertEqual(sorted(os.listdir(subject_dir)),
                         [BasicRepo.PROV_DIR, 'subject_sink.txt'])
        visit_dir = self.get_session_dir(
            frequency='per_visit',
            from_study=self.SUMMARY_STUDY_NAME)
        self.assertEqual(sorted(os.listdir(visit_dir)),
                         [BasicRepo.PROV_DIR, 'visit_sink.txt'])
        project_dir = self.get_session_dir(
            frequency='per_study',
            from_study=self.SUMMARY_STUDY_NAME)
        self.assertEqual(sorted(os.listdir(project_dir)),
                         [BasicRepo.PROV_DIR, 'study_sink.txt'])
        # Reload the data from the summary directories
        reloadinputnode = pe.Node(
            IdentityInterface(['subject_id', 'visit_id']),
            name='reload_inputnode')
        reloadinputnode.inputs.subject_id = self.SUBJECT
        reloadinputnode.inputs.visit_id = self.VISIT
        reloadsource_per_subject = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in subject_sink_files),
            name='reload_source_per_subject')
        reloadsource_per_visit = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in visit_sink_files),
            name='reload_source_per_visit')
        reloadsource_per_study = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in study_sink_files),
            name='reload_source_per_study')
        reloadsink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection
                 for f in ['resink1', 'resink2', 'resink3']),
                dummy_pipeline),
            name='reload_sink')
        reloadsink.inputs.name = 'reload_summary'
        reloadsink.inputs.desc = (
            "Tests the reloading of subject and project summary filesets")
        reloadworkflow = pe.Workflow('reload_summary_unittest',
                                     base_dir=self.work_dir)
        for node in (reloadsource_per_subject, reloadsource_per_visit,
                     reloadsource_per_study, reloadsink):
            for iterator in ('subject_id', 'visit_id'):
                reloadworkflow.connect(reloadinputnode, iterator,
                                       node, iterator)
        reloadworkflow.connect(reloadsource_per_subject,
                               'subject_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink1' + PATH_SUFFIX)
        reloadworkflow.connect(reloadsource_per_visit,
                               'visit_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink2' + PATH_SUFFIX)
        reloadworkflow.connect(reloadsource_per_study,
                               'study_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink3' + PATH_SUFFIX)
        reloadworkflow.run()
        outputs = [
            f for f in sorted(os.listdir(
                self.get_session_dir(from_study=self.SUMMARY_STUDY_NAME)))
            if f not in (BasicRepo.FIELDS_FNAME,
                         BasicRepo.PROV_DIR)]
        self.assertEqual(outputs,
                         ['resink1.txt', 'resink2.txt', 'resink3.txt'])



class ArcanaException(Exception):

    @property
    def msg(self):
        return self.args[0]

    @msg.setter
    def msg(self, msg):
        self.args = (msg,) + self.args[1:]


class ArcanaError(ArcanaException):
    pass


class ArcanaNotBoundToStudyError(ArcanaError):
    pass


class ArcanaVersionError(ArcanaError):
    pass


class ArcanaRequirementNotFoundError(ArcanaVersionError):
    pass


class ArcanaVersionNotDetectableError(ArcanaVersionError):
    pass


class ArcanaEnvModuleNotLoadedError(ArcanaError):
    pass


class ArcanaMissingInputError(ArcanaException):
    pass


class ArcanaProtectedOutputConflictError(ArcanaError):
    pass


class ArcanaCantPickleStudyError(ArcanaError):
    pass


class ArcanaRepositoryError(ArcanaError):
    pass


class ArcanaUsageError(ArcanaError):
    pass


class ArcanaDesignError(ArcanaError):
    pass


class NamedArcanaError(ArcanaError):

    def __init__(self, name, msg):
        super(NamedArcanaError, self).__init__(msg)
        self.name = name


class ArcanaNameError(NamedArcanaError):
    pass


class ArcanaIndexError(ArcanaError):

    def __init__(self, index, msg):
        super(ArcanaIndexError, self).__init__(msg)
        self.index = index


class ArcanaMissingDataException(ArcanaUsageError):
    pass


class ArcanaDataNotDerivedYetError(NamedArcanaError, ArcanaDesignError):
    pass


class ArcanaInputError(ArcanaUsageError):
    pass


class ArcanaInputMissingMatchError(ArcanaInputError):
    pass


class ArcanaOutputNotProducedException(ArcanaException):
    """
    Raised when a given spec is not produced due to switches and inputs
    provided to the study
    """


class ArcanaInsufficientRepoDepthError(ArcanaError):
    pass


class ArcanaFileFormatError(ArcanaError):
    pass


class ArcanaFilesetNotCachedException(ArcanaException):
    pass


class NoMatchingPipelineException(ArcanaException):
    pass


class ArcanaModulesError(ArcanaError):
    pass


class ArcanaModulesNotInstalledException(ArcanaException):
    pass


class ArcanaJobSubmittedException(ArcanaException):
    """
    Signifies that a pipeline has been submitted to a scheduler and
    a return value won't be returned.
    """


class ArcanaNoRunRequiredException(ArcanaException):
    """
    Used to signify when a pipeline doesn't need to be run as all
    required outputs are already present in the repository
    """


class ArcanaFileFormatClashError(ArcanaError):
    """
    Used when two mismatching data formats are registered with the same
    name or extention
    """


class ArcanaNoConverterError(ArcanaError):
    "No converters exist between formats"


class ArcanaConverterNotAvailableError(ArcanaError):
    "The converter required to convert between formats is not "
    "available"


class ArcanaReprocessException(ArcanaException):
    pass


class ArcanaWrongRepositoryError(ArcanaError):
    pass


class ArcanaIvalidParameterError(ArcanaError):
    pass



from builtins import zip
from builtins import object
from itertools import chain, groupby
from collections import defaultdict
from operator import attrgetter, itemgetter
from collections import OrderedDict
import logging
from arcana.data import BaseFileset, BaseField
from arcana.utils import split_extension
from arcana.exceptions import (
    ArcanaNameError, ArcanaRepositoryError)

id_getter = attrgetter('id')

logger = logging.getLogger('arcana')


class TreeNode(object):

    def __init__(self, filesets, fields, records):
        if filesets is None:
            filesets = []
        if fields is None:
            fields = []
        if records is None:
            records = []
        # Save filesets and fields in ordered dictionary by name and
        # name of study that generated them (if applicable)
        self._filesets = OrderedDict()
        for fileset in sorted(filesets):
            id_key = (fileset.id, fileset.from_study)
            try:
                dct = self._filesets[id_key]
            except KeyError:
                dct = self._filesets[id_key] = OrderedDict()
            if fileset.format_name is not None:
                format_key = fileset.format_name
            else:
                format_key = split_extension(fileset.path)[1]
            if format_key in dct:
                raise ArcanaRepositoryError(
                    "Attempting to add duplicate filesets to tree ({} and {})"
                    .format(fileset, dct[format_key]))
            dct[format_key] = fileset
        self._fields = OrderedDict(((f.name, f.from_study), f)
                                   for f in sorted(fields))
        self._records = OrderedDict(
            ((r.pipeline_name, r.from_study), r)
            for r in sorted(records, key=lambda r: (r.subject_id, r.visit_id,
                                                    r.from_study)))
        self._missing_records = []
        self._duplicate_records = []
        # Match up provenance records with items in the node
        for item in chain(self.filesets, self.fields):
            if not item.derived:
                continue  # Skip acquired items
            records = [r for r in self.records
                       if (item.from_study == r.from_study and
                           item.name in r.outputs)]
            if not records:
                self._missing_records.append(item.name)
            elif len(records) > 1:
                item.record = sorted(records, key=attrgetter('datetime'))[-1]
                self._duplicate_records.append(item.name)
            else:
                item.record = records[0]

    def __eq__(self, other):
        if not (isinstance(other, type(self)) or
                isinstance(self, type(other))):
            return False
        return (tuple(self.filesets) == tuple(other.filesets) and
                tuple(self.fields) == tuple(other.fields) and
                tuple(self.records) == tuple(other.records))

    def __hash__(self):
        return (hash(tuple(self.filesets)) ^ hash(tuple(self.fields)) ^
                hash(tuple(self.fields)))

    @property
    def filesets(self):
        return chain(*(d.values() for d in self._filesets.values()))

    @property
    def fields(self):
        return self._fields.values()

    @property
    def records(self):
        return self._records.values()

    @property
    def subject_id(self):
        "To be overridden by subclasses where appropriate"
        return None

    @property
    def visit_id(self):
        "To be overridden by subclasses where appropriate"
        return None

    def fileset(self, id, from_study=None, format=None):  # @ReservedAssignment @IgnorePep8
        """
        Gets the fileset with the ID 'id' produced by the Study named 'study'
        if provided. If a spec is passed instead of a str to the name argument,
        then the study will be set from the spec iff it is derived

        Parameters
        ----------
        id : str | FilesetSpec
            The name of the fileset or a spec matching the given name
        from_study : str | None
            Name of the study that produced the fileset if derived. If None
            and a spec is passed instaed of string to the name argument then
            the study name will be taken from the spec instead.
        format : FileFormat | str | None
            Either the format of the fileset to return or the name of the
            format. If None and only a single fileset is found for the given
            name and study then that is returned otherwise an exception is
            raised
        """
        if isinstance(id, BaseFileset):
            if from_study is None and id.derived:
                from_study = id.study.name
            id = id.name  # @ReservedAssignment
        try:
            format_dct = self._filesets[(id, from_study)]
        except KeyError:
            available = [
                ('{}(format={})'.format(f.id, f._resource_name)
                 if f._resource_name is not None else f.id)
                for f in self.filesets if f.from_study == from_study]
            other_studies = [
                (f.from_study if f.from_study is not None else '<root>')
                for f in self.filesets if f.id == id]
            if other_studies:
                msg = (". NB: matching fileset(s) found for '{}' study(ies) "
                       "('{}')".format(id, "', '".join(other_studies)))
            else:
                msg = ''
            raise ArcanaNameError(
                id,
                ("{} doesn't have a fileset named '{}'{} "
                   "(available '{}'){}"
                   .format(self, id,
                           (" from study '{}'".format(from_study)
                            if from_study is not None else ''),
                           "', '".join(available), msg)))
        else:
            if format is None:
                all_formats = list(format_dct.values())
                if len(all_formats) > 1:
                    raise ArcanaNameError(
                        "Multiple filesets found for '{}'{} in {} with formats"
                        " {}. Need to specify a format"
                        .format(id, ("in '{}'".format(from_study)
                                       if from_study is not None else ''),
                                self, "', '".join(format_dct.keys())))
                fileset = all_formats[0]
            else:
                try:
                    if isinstance(format, str):
                        fileset = format_dct[format]
                    else:
                        try:
                            fileset = format_dct[format.ext]
                        except KeyError:
                            fileset = None
                            for rname, rfileset in format_dct.items():
                                if rname in format.resource_names(
                                        self.tree.repository.type):
                                    fileset = rfileset
                                    break
                            if fileset is None:
                                raise
                except KeyError:
                    raise ArcanaNameError(
                        format,
                        ("{} doesn't have a fileset named '{}'{} with "
                         "format '{}' (available '{}'){}"
                           .format(self, id,
                                   (" from study '{}'".format(from_study)
                                    if from_study is not None else ''),
                                   format,
                                   "', '".join(format_dct.keys()), msg)))

        return fileset

    def field(self, name, from_study=None):
        """
        Gets the field named 'name' produced by the Study named 'study' if
        provided. If a spec is passed instead of a str to the name argument,
        then the study will be set from the spec iff it is derived

        Parameters
        ----------
        name : str | BaseField
            The name of the field or a spec matching the given name
        study : str | None
            Name of the study that produced the field if derived. If None
            and a spec is passed instaed of string to the name argument then
            the study name will be taken from the spec instead.
        """
        if isinstance(name, BaseField):
            if from_study is None and name.derived:
                from_study = name.study.name
            name = name.name
        try:
            return self._fields[(name, from_study)]
        except KeyError:
            available = [d.name for d in self.fields
                         if d.from_study == from_study]
            other_studies = [(d.from_study if d.from_study is not None
                              else '<root>')
                             for d in self.fields
                             if d.name == name]
            if other_studies:
                msg = (". NB: matching field(s) found for '{}' study(ies) "
                       "('{}')".format(name, "', '".join(other_studies)))
            else:
                msg = ''
            raise ArcanaNameError(
                name, ("{} doesn't have a field named '{}'{} "
                       "(available '{}')"
                       .format(
                           self, name,
                           (" from study '{}'".format(from_study)
                            if from_study is not None else ''),
                           "', '".join(available), msg)))

    def record(self, pipeline_name, from_study):
        """
        Returns the provenance record for a given pipeline

        Parameters
        ----------
        pipeline_name : str
            The name of the pipeline that generated the record
        from_study : str
            The name of the study that the pipeline was generated from

        Returns
        -------
        record : arcana.provenance.Record
            The provenance record generated by the specified pipeline
        """
        try:
            return self._records[(pipeline_name, from_study)]
        except KeyError:
            found = []
            for sname, pnames in groupby(sorted(self._records,
                                                key=itemgetter(1)),
                                         key=itemgetter(1)):
                found.append(
                    "'{}' for '{}'".format("', '".join(p for p, _ in pnames),
                                           sname))
            raise ArcanaNameError(
                (pipeline_name, from_study),
                ("{} doesn't have a provenance record for pipeline '{}' "
                 "for '{}' study (found {})".format(
                     self, pipeline_name, from_study,
                     '; '.join(found))))

    @property
    def data(self):
        return chain(self.filesets, self.fields)

    def __ne__(self, other):
        return not (self == other)

    def find_mismatch(self, other, indent=''):
        """
        Highlights where two nodes differ in a human-readable form

        Parameters
        ----------
        other : TreeNode
            The node to compare
        indent : str
            The white-space with which to indent output string

        Returns
        -------
        mismatch : str
            The human-readable mismatch string
        """
        if self != other:
            mismatch = "\n{}{}".format(indent, type(self).__name__)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if len(list(self.filesets)) != len(list(other.filesets)):
            mismatch += ('\n{indent}mismatching summary fileset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.filesets)),
                                 len(list(other.filesets)),
                                 list(self.filesets),
                                 list(other.filesets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.filesets, other.filesets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching summary field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch


class Tree(TreeNode):
    """
    Represents a project tree as stored in a repository

    Parameters
    ----------
    subjects : List[Subject]
        List of subjects
    visits : List[Visits]
        List of visits in the project across subjects
        (i.e. timepoint 1, 2, 3)
    repository : Repository
        The repository that the tree comes from
    filesets : List[Fileset]
        The filesets that belong to the project, i.e. of 'per_study'
        frequency
    fields : List[Field]
        The fields that belong to the project, i.e. of 'per_study'
        frequency
    fill_subjects : list[int] | None
        Create empty sessions for any subjects that are missing
        from the provided list. Typically only used if all
        the inputs to the study are coming from different repositories
        to the one that the derived products are stored in
    fill_visits : list[int] | None
        Create empty sessions for any visits that are missing
        from the provided list. Typically only used if all
        the inputs to the study are coming from different repositories
        to the one that the derived products are stored in
    """

    frequency = 'per_study'

    def __init__(self, subjects, visits, repository, filesets=None,
                 fields=None, records=None, fill_subjects=None,
                 fill_visits=None, **kwargs):  # @UnusedVariable @IgnorePep8
        TreeNode.__init__(self, filesets, fields, records)
        self._subjects = OrderedDict(sorted(
            ((s.id, s) for s in subjects), key=itemgetter(0)))
        self._visits = OrderedDict(sorted(
            ((v.id, v) for v in visits), key=itemgetter(0)))
        if fill_subjects is not None or fill_visits is not None:
            self._fill_empty_sessions(fill_subjects, fill_visits)
        for subject in self.subjects:
            subject.tree = self
        for visit in self.visits:
            visit.tree = self
        for session in self.sessions:
            session.tree = self
        self._repository = repository
        # Collate missing and duplicates provenance records for single warnings
        missing_records = defaultdict(lambda: defaultdict(list))
        duplicate_records = defaultdict(lambda: defaultdict(list))
        for node in self.nodes():
            for missing in node._missing_records:
                missing_records[missing][node.visit_id].append(node.subject_id)
            for duplicate in node._duplicate_records:
                duplicate_records[duplicate][node.visit_id].append(
                    node.subject_id)
        for name, ids in missing_records.items():
            logger.warning(
                "No provenance records found for {} derivative in "
                "the following nodes: {}. Will assume they are a "
                "\"protected\" (manually created) derivatives"
                .format(name, '; '.join("visit='{}', subjects={}".format(k, v)
                                        for k, v in ids.items())))
        for name, ids in duplicate_records.items():
            logger.warning(
                "Duplicate provenance records found for {} in the following "
                "nodes: {}. Will select the latest record in each case"
                .format(name, '; '.join("visit='{}', subjects={}".format(k, v)
                                        for k, v in ids.items())))

    def __eq__(self, other):
        return (super(Tree, self).__eq__(other) and
                self._subjects == other._subjects and
                self._visits == other._visits)

    def __hash__(self):
        return (TreeNode.__hash_(self) ^
                hash(tuple(self.subjects)) ^
                hash(tuple(self._visits)))

    @property
    def repository(self):
        return self._repository

    @property
    def subjects(self):
        return self._subjects.values()

    @property
    def visits(self):
        return self._visits.values()

    @property
    def sessions(self):
        return chain(*(s.sessions for s in self.subjects))

    @property
    def tree(self):
        return self

    @property
    def subject_ids(self):
        return self._subjects.keys()

    @property
    def visit_ids(self):
        return self._visits.keys()

    @property
    def session_ids(self):
        return ((s.subject_id, s.visit_id) for s in self.sessions)

    @property
    def complete_subjects(self):
        max_num_sessions = max(len(s) for s in self.subjects)
        return (s for s in self.subjects if len(s) == max_num_sessions)

    @property
    def complete_visits(self):
        max_num_sessions = max(len(v) for v in self.visits)
        return (v for v in self.visits if len(v) == max_num_sessions)

    @property
    def incomplete_subjects(self):
        max_num_sessions = max(len(s) for s in self.subjects)
        return (s for s in self.subjects if len(s) != max_num_sessions)

    @property
    def incomplete_visits(self):
        max_num_sessions = max(len(v) for v in self.visits)
        return (v for v in self.visits if len(v) != max_num_sessions)

    def subject(self, id):  # @ReservedAssignment
        try:
            return self._subjects[str(id)]
        except KeyError:
            raise ArcanaNameError(
                id, ("{} doesn't have a subject named '{}' ('{}')"
                       .format(self, id, "', '".join(self._subjects))))

    def visit(self, id):  # @ReservedAssignment
        try:
            return self._visits[str(id)]
        except KeyError:
            raise ArcanaNameError(
                id, ("{} doesn't have a visit named '{}' ('{}')"
                       .format(self, id, "', '".join(self._visits))))

    def session(self, subject_id, visit_id):
        return self.subject(subject_id).session(visit_id)

    def __iter__(self):
        return self.nodes()

    def nodes(self, frequency=None):
        """
        Returns an iterator over all nodes in the tree for the specified
        frequency. If no frequency is specified then all nodes are returned

        Parameters
        ----------
        frequency : str | None
            The frequency of the nodes to iterate over. If None all
            frequencies are returned

        Returns
        -------
        nodes : iterable[TreeNode]
        """
        if frequency is None:
            nodes = chain(*(self._nodes(f)
                            for f in ('per_study', 'per_subject',
                                      'per_visit', 'per_session')))
        else:
            nodes = self._nodes(frequency=frequency)
        return nodes

    def _nodes(self, frequency):
        if frequency == 'per_session':
            nodes = chain(*(s.sessions for s in self.subjects))
        elif frequency == 'per_subject':
            nodes = self.subjects
        elif frequency == 'per_visit':
            nodes = self.visits
        elif frequency == 'per_study':
            nodes = [self]
        else:
            assert False
        return nodes

    def find_mismatch(self, other, indent=''):
        """
        Used in debugging unittests
        """
        mismatch = super(Tree, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if len(list(self.subjects)) != len(list(other.subjects)):
            mismatch += ('\n{indent}mismatching subject lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.subjects)),
                                 len(list(other.subjects)),
                                 list(self.subjects),
                                 list(other.subjects),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.subjects, other.subjects):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.visits)) != len(list(other.visits)):
            mismatch += ('\n{indent}mismatching visit lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.visits)),
                                 len(list(other.visits)),
                                 list(self.visits),
                                 list(other.visits),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.visits, other.visits):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __repr__(self):
        return ("Tree(num_subjects={}, num_visits={}, "
                "num_filesets={}, num_fields={})".format(
                    len(list(self.subjects)),
                    len(list(self.visits)),
                    len(list(self.filesets)), len(list(self.fields))))

    def _fill_empty_sessions(self, fill_subjects, fill_visits):
        """
        Fill in tree with additional empty subjects and/or visits to
        allow the study to pull its inputs from external repositories
        """
        if fill_subjects is None:
            fill_subjects = [s.id for s in self.subjects]
        if fill_visits is None:
            fill_visits = [v.id for v in self.complete_visits]
        for subject_id in fill_subjects:
            try:
                subject = self.subject(subject_id)
            except ArcanaNameError:
                subject = self._subjects[subject_id] = Subject(
                    subject_id, [], [], [])
            for visit_id in fill_visits:
                try:
                    subject.session(visit_id)
                except ArcanaNameError:
                    session = Session(subject_id, visit_id, [], [])
                    subject._sessions[visit_id] = session
                    try:
                        visit = self.visit(visit_id)
                    except ArcanaNameError:
                        visit = self._visits[visit_id] = Visit(
                            visit_id, [], [], [])
                    visit._sessions[subject_id] = session

    @classmethod
    def construct(cls, repository, filesets=(), fields=(), records=(),
                  file_formats=(), **kwargs):
        """
        Return the hierarchical tree of the filesets and fields stored in a
        repository

        Parameters
        ----------
        respository : Repository
            The repository that the tree comes from
        filesets : list[Fileset]
            List of all filesets in the tree
        fields : list[Field]
            List of all fields in the tree
        records : list[Record]
            List of all records in the tree

        Returns
        -------
        tree : arcana.repository.Tree
            A hierarchical tree of subject, session and fileset
            information for the repository
        """
        # Sort the data by subject and visit ID
        filesets_dict = defaultdict(list)
        for fset in filesets:
            if file_formats:
                fset.set_format(file_formats)
            filesets_dict[(fset.subject_id, fset.visit_id)].append(fset)
        fields_dict = defaultdict(list)
        for field in fields:
            fields_dict[(field.subject_id, field.visit_id)].append(field)
        records_dict = defaultdict(list)
        for record in records:
            records_dict[(record.subject_id, record.visit_id)].append(record)
        # Create all sessions
        subj_sessions = defaultdict(list)
        visit_sessions = defaultdict(list)
        for sess_id in set(chain(filesets_dict, fields_dict,
                                 records_dict)):
            if None in sess_id:
                continue  # Save summaries for later
            subj_id, visit_id = sess_id
            session = Session(
                subject_id=subj_id, visit_id=visit_id,
                filesets=filesets_dict[sess_id],
                fields=fields_dict[sess_id],
                records=records_dict[sess_id])
            subj_sessions[subj_id].append(session)
            visit_sessions[visit_id].append(session)
        subjects = []
        for subj_id in subj_sessions:
            subjects.append(Subject(
                subj_id,
                sorted(subj_sessions[subj_id]),
                filesets_dict[(subj_id, None)],
                fields_dict[(subj_id, None)],
                records_dict[(subj_id, None)]))
        visits = []
        for visit_id in visit_sessions:
            visits.append(Visit(
                visit_id,
                sorted(visit_sessions[visit_id]),
                filesets_dict[(None, visit_id)],
                fields_dict[(None, visit_id)],
                records_dict[(None, visit_id)]))
        return Tree(sorted(subjects),
                    sorted(visits),
                    repository,
                    filesets_dict[(None, None)],
                    fields_dict[(None, None)],
                    records_dict[(None, None)],
                    **kwargs)


class Subject(TreeNode):
    """
    Represents a subject as stored in a repository

    Parameters
    ----------
    subject_id : str
        The ID of the subject
    sessions : List[Session]
        The sessions in the subject
    filesets : List[Fileset]
        The filesets that belong to the subject, i.e. of 'per_subject'
        frequency
    fields : List[Field]
        The fields that belong to the subject, i.e. of 'per_subject'
        frequency
    """

    frequency = 'per_subject'

    def __init__(self, subject_id, sessions, filesets=None,
                 fields=None, records=None):
        TreeNode.__init__(self, filesets, fields, records)
        self._id = subject_id
        self._sessions = OrderedDict(sorted(
            ((s.visit_id, s) for s in sessions), key=itemgetter(0)))
        for session in self.sessions:
            session.subject = self
        self._tree = None

    @property
    def id(self):
        return self._id

    @property
    def subject_id(self):
        return self.id

    @property
    def tree(self):
        return self._tree

    @tree.setter
    def tree(self, tree):
        self._tree = tree

    def __lt__(self, other):
        return self._id < other._id

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other)and
                self._id == other._id and
                self._sessions == other._sessions)

    def __hash__(self):
        return (TreeNode.__hash__(self) ^
                hash(self._id) ^
                hash(tuple(self.sessions)))

    def __len__(self):
        return len(self._sessions)

    def __iter__(self):
        return self.sessions

    @property
    def sessions(self):
        return self._sessions.values()

    def nodes(self, frequency=None):
        """
        Returns all sessions in the subject. If a frequency is passed then
        it will return all nodes of that frequency related to the current node.
        If there is no relationshop between the current node and the frequency
        then all nodes in the tree for that frequency will be returned

        Parameters
        ----------
        frequency : str | None
            The frequency of the nodes to return

        Returns
        -------
        nodes : iterable[TreeNode]
            All nodes related to the subject for the specified frequency, or
            all nodes in the tree if there is no relation with that frequency
            (e.g. per_visit)
        """
        if frequency in (None, 'per_session'):
            return self.sessions
        elif frequency == 'per_visit':
            return self.parent.nodes(frequency)
        elif frequency == 'per_subject':
            return [self]
        elif frequency == 'per_study':
            return [self.parent]

    @property
    def visit_ids(self):
        return self._sessions.values()

    def session(self, visit_id):
        try:
            return self._sessions[str(visit_id)]
        except KeyError:
            raise ArcanaNameError(
                visit_id, ("{} doesn't have a session named '{}' ('{}')"
                           .format(self, visit_id,
                                   "', '".join(self._sessions))))

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Subject(id={}, num_sessions={})"
                .format(self._id, len(self._sessions)))


class Visit(TreeNode):
    """
    Represents a collection of visits across subjects (e.g. time-point 1)
    as stored in a repository

    Parameters
    ----------
    visit_id : str
        The ID of the visit
    sessions : List[Session]
        The sessions in the visit
    filesets : List[Fileset]
        The filesets that belong to the visit, i.e. of 'per_visit'
        frequency
    fields : List[Field]
        The fields that belong to the visit, i.e. of 'per_visit'
        frequency
    """

    frequency = 'per_visit'

    def __init__(self, visit_id, sessions, filesets=None, fields=None,
                 records=None):
        TreeNode.__init__(self, filesets, fields, records)
        self._id = visit_id
        self._sessions = OrderedDict(sorted(
            ((s.subject_id, s) for s in sessions), key=itemgetter(0)))
        for session in sessions:
            session.visit = self
        self._tree = None

    @property
    def id(self):
        return self._id

    @property
    def visit_id(self):
        return self.id

    @property
    def tree(self):
        return self._tree

    @tree.setter
    def tree(self, tree):
        self._tree = tree

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other) and
                self._id == other._id and
                self._sessions == other._sessions)

    def __hash__(self):
        return (TreeNode.__hash__(self) ^
                hash(self._id) ^
                hash(tuple(self.sessions)))

    def __lt__(self, other):
        return self._id < other._id

    def __len__(self):
        return len(self._sessions)

    def __iter__(self):
        return self.sessions

    @property
    def sessions(self):
        return self._sessions.values()

    def nodes(self, frequency=None):
        """
        Returns all sessions in the visit. If a frequency is passed then
        it will return all nodes of that frequency related to the current node.
        If there is no relationshop between the current node and the frequency
        then all nodes in the tree for that frequency will be returned

        Parameters
        ----------
        frequency : str | None
            The frequency of the nodes to return

        Returns
        -------
        nodes : iterable[TreeNode]
            All nodes related to the visit for the specified frequency, or
            all nodes in the tree if there is no relation with that frequency
            (e.g. per_subject)
        """
        if frequency in (None, 'per_session'):
            return self.sessions
        elif frequency == 'per_subject':
            return self.parent.nodes(frequency)
        elif frequency == 'per_visit':
            return [self]
        elif frequency == 'per_study':
            return [self.parent]

    def session(self, subject_id):
        try:
            return self._sessions[str(subject_id)]
        except KeyError:
            raise ArcanaNameError(
                subject_id, ("{} doesn't have a session named '{}' ('{}')"
                             .format(self, subject_id,
                                     "', '".join(self._sessions))))

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "Visit(id={}, num_sessions={})".format(self._id,
                                                      len(self._sessions))


class Session(TreeNode):
    """
    Represents a session stored in a repository

    Parameters
    ----------
    subject_id : str
        The subject ID of the session
    visit_id : str
        The visit ID of the session
    filesets : list(Fileset)
        The filesets found in the session
    derived : dict[str, Session]
        Sessions storing derived scans are stored for separate analyses
    """

    frequency = 'per_session'

    def __init__(self, subject_id, visit_id, filesets=None, fields=None,
                 records=None):
        TreeNode.__init__(self, filesets, fields, records)
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._subject = None
        self._visit = None
        self._tree = None

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def subject_id(self):
        return self._subject_id

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other) and
                self.subject_id == other.subject_id and
                self.visit_id == other.visit_id)

    def __hash__(self):
        return (TreeNode.__hash__(self) ^
                hash(self.subject_id) ^
                hash(self.visit_id))

    def __lt__(self, other):
        if self.subject_id < other.subject_id:
            return True
        else:
            return self.visit_id < other.visit_id

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def visit(self):
        return self._visit

    @visit.setter
    def visit(self, visit):
        self._visit = visit

    @property
    def tree(self):
        return self._tree

    @tree.setter
    def tree(self, tree):
        self._tree = tree

    def nodes(self, frequency=None):
        """
        Returns all nodes of the specified frequency that are related to
        the given Session

        Parameters
        ----------
        frequency : str | None
            The frequency of the nodes to return

        Returns
        -------
        nodes : iterable[TreeNode]
            All nodes related to the Session for the specified frequency
        """
        if frequency is None:
            []
        elif frequency == 'per_session':
            return [self]
        elif frequency in ('per_visit', 'per_subject'):
            return [self.parent]
        elif frequency == 'per_study':
            return [self.parent.parent]

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Session(subject_id='{}', visit_id='{}', num_filesets={}, "
                "num_fields={})".format(
                    self.subject_id, self.visit_id, len(self._filesets),
                    len(self._fields)))



from __future__ import division
from builtins import str  # @UnusedImports
from builtins import object
import time
import logging
from nipype.interfaces.base import isdefined
from nipype.pipeline.engine import (
    Node as NipypeNode, JoinNode as NipypeJoinNode,
    MapNode as NipypeMapNode)
from arcana.utils import get_class_info, HOSTNAME


logger = logging.getLogger('arcana')


class NodeMixin(object):
    """
    A Mixin class used to insert additional meta data into the node (than
    the Nipype node it is used to subclass), and report accuracy of wall_time
    estimates

    Parameters
    ----------
    environment : Environment | None
        The environment within which to execute the node (automatically added
        by Arcana)
    requirements : list(Version | VersionRange)
        List of minimum software versions or range of software versions
        required
    wall_time : int
        Expected wall time of the node in minutes.
    annotations : dict[str, *]
        Flexible annotations that can be used to optimise how the node is
        executed by the processor (e.g. whether GPU cards are required)
    """

    def __init__(self, environment, *args, **kwargs):
        self._environment = environment
        # Get versions of software in the environment that satisfy the given
        # requirements
        requirements = kwargs.pop('requirements', [])
        self._versions = self._environment.satisfy(*requirements)
        self._wall_time = kwargs.pop('wall_time', None)
        self._annotations = kwargs.pop('annotations', {})
        self.nipype_cls.__init__(self, *args, **kwargs)

    def _run_command(self, *args, **kwargs):
        # Detect run time and compare against specified wall_time
        start_time = time.time()
        result = self.nipype_cls._run_command(self, *args, **kwargs)
        end_time = time.time()
        run_time = (end_time - start_time) // 60
        if run_time > self._wall_time:
            logger.warning("Executed '{}' node in {} minutes, which is longer "
                           "than specified wall time ({} minutes)"
                           .format(self.name, run_time,
                                   self._wall_time))
        else:
            logger.info("Executed '{}' node in {} minutes"
                        .format(self.name, run_time))
        return result

    @property
    def annotations(self):
        return self._annotations

    @property
    def versions(self):
        return self._versions

    @property
    def environment(self):
        return self._environment

    @property
    def prov(self):
        prov = {
            'interface': get_class_info(type(self.interface)),
            'requirements': {v.name: v.prov for v in self.versions},
            'parameters': {}}
        for trait_name in self.inputs.visible_traits():
            val = getattr(self.inputs, trait_name)
            try:
                val_prov = val.prov
            except AttributeError:
                val_prov = val
            if isdefined(val):
                prov['parameters'][trait_name] = val_prov
        return prov


class Node(NodeMixin, NipypeNode):

    nipype_cls = NipypeNode  # Not req. in Py3 where super() in mixin works


class JoinNode(NodeMixin, NipypeJoinNode):

    nipype_cls = NipypeJoinNode  # Not req. in Py3 where super() in mixin works


class MapNode(NodeMixin, NipypeMapNode):

    node_cls = Node
    nipype_cls = NipypeMapNode  # Not req. in Py3 where super() in mixin works

    def _make_nodes(self, cwd=None):
        """
        Cast generated nodes to be Arcana nodes
        """
        for i, node in NipypeMapNode._make_nodes(self, cwd=cwd):
            # "Cast" NiPype node to a Arcana Node and set Arcana Node
            # parameters
            node.__class__ = self.node_cls
            node._environment = self._environment
            node._versions = self._versions
            node._wall_time = self._wall_time
            node._annotations = self._annotations
            yield i, node


class Environment(object):
    """
    Base class for all Environment classes
    """

    node_types = {'base': Node, 'map': MapNode, 'join': JoinNode}

    def satisfy(self, *requirements):
        """
        Checks whether the given requirements are satisfiable within the given
        execution context and returns the corresponding versions that will be
        used.

        Parameter
        ---------
        requirements : list(Requirement)
            List of requirements to check whether they are satisfiable

        Returns
        -------
        versions : list(Version)
            Exact software versions to be executed in the environment
        """
        raise NotImplementedError

    @property
    def prov(self):
        return {
            'type': get_class_info(type(self)),
            'host': HOSTNAME}



from __future__ import absolute_import
import os.path as op
import tempfile
import unittest
import sys
from collections import OrderedDict
from unittest import TestCase
import xnat
from arcana.utils.testing import BaseTestCase
from arcana.data import InputFilesets
from arcana.data.file_format import text_format
from arcana.repository import XnatRepo
from arcana.processor import SingleProc
from arcana.utils.testing.xnat import SKIP_ARGS, SERVER, TestOnXnatMixin


# Import TestExistingPrereqs study to test it on XNAT
sys.path.insert(0, op.join(op.dirname(__file__), '..', '..'))
import test_data  # @UnresolvedImport @IgnorePep8
from test_data import dicom_format  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)

# Import test_local to run TestProjectInfo on XNAT using TestOnXnat mixin
sys.path.insert(0, op.join(op.dirname(__file__), '..', '..', 'processor'))
import test_to_process  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)


class TestConnectDisconnect(TestCase):

    @unittest.skipIf(*SKIP_ARGS)
    def test_connect_disconnect(self):
        repository = XnatRepo(project_id='dummy',
                              server=SERVER,
                              cache_dir=tempfile.mkdtemp())
        with repository:
            self._test_open(repository)
        self._test_closed(repository)

        with repository:
            self._test_open(repository)
            with repository:
                self._test_open(repository)
            self._test_open(repository)
        self._test_closed(repository)

    def _test_open(self, repository):
        repository._login.classes  # check connection

    def _test_closed(self, repository):
        self.assertRaises(
            AttributeError,
            getattr,
            repository._login,
            'classes')


class TestProvInputChangeOnXnat(TestOnXnatMixin,
                                test_to_process.TestProvInputChange):

    BASE_CLASS = test_to_process.TestProvInputChange

    @unittest.skipIf(*SKIP_ARGS)
    def test_input_change(self):
        super(TestProvInputChangeOnXnat, self).test_input_change()


class TestDicomTagMatchAndIDOnXnat(TestOnXnatMixin,
                                   test_data.TestDicomTagMatch):

    BASE_CLASS = test_data.TestDicomTagMatch
    REF_FORMATS = [dicom_format]

    @property
    def ref_dir(self):
        return op.join(
            self.ref_path, self._get_name(self.BASE_CLASS))

    def setUp(self):
        test_data.TestDicomTagMatch.setUp(self)
        TestOnXnatMixin.setUp(self)
        # Set up DICOM headers
        with xnat.connect(SERVER) as login:
            xsess = login.projects[self.project].experiments[
                '_'.join((self.project, self.SUBJECT, self.VISIT))]
            login.put('/data/experiments/{}?pullDataFromHeaders=true'
                      .format(xsess.id))

    def tearDown(self):
        TestOnXnatMixin.tearDown(self)
        test_data.TestDicomTagMatch.tearDown(self)

    @unittest.skipIf(*SKIP_ARGS)
    def test_dicom_match(self):
        study = test_data.TestMatchStudy(
            name='test_dicom',
            repository=XnatRepo(
                project_id=self.project,
                server=SERVER, cache_dir=tempfile.mkdtemp()),
            processor=SingleProc(self.work_dir),
            inputs=test_data.TestDicomTagMatch.DICOM_MATCH)
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')

    @unittest.skipIf(*SKIP_ARGS)
    def test_id_match(self):
        study = test_data.TestMatchStudy(
            name='test_dicom',
            repository=XnatRepo(
                project_id=self.project,
                server=SERVER, cache_dir=tempfile.mkdtemp()),
            processor=SingleProc(self.work_dir),
            inputs=[
                InputFilesets('gre_phase', valid_formats=dicom_format, id=7),
                InputFilesets('gre_mag', valid_formats=dicom_format, id=6)])
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')

    @unittest.skipIf(*SKIP_ARGS)
    def test_order_match(self):
        test_data.TestDicomTagMatch.test_order_match(self)


class TestFilesetCacheOnPathAccess(TestOnXnatMixin, BaseTestCase):

    INPUT_FILESETS = {'fileset': '1'}

    @unittest.skipIf(*SKIP_ARGS)
    def test_cache_on_path_access(self):
        tmp_dir = tempfile.mkdtemp()
        repository = XnatRepo(
            project_id=self.project,
            server=SERVER, cache_dir=tmp_dir)
        tree = repository.tree(
            subject_ids=[self.SUBJECT],
            visit_ids=[self.VISIT])
        # Get a fileset
        fileset = list(list(list(tree.subjects)[0].sessions)[0].filesets)[0]
        fileset.format = text_format
        self.assertEqual(fileset._path, None)
        target_path = op.join(
            tmp_dir, self.project,
            '{}_{}'.format(self.project, self.SUBJECT),
            '{}_{}_{}'.format(self.project, self.SUBJECT, self.VISIT),
            fileset.basename, fileset.fname)
        # This should implicitly download the fileset
        self.assertEqual(fileset.path, target_path)
        with open(target_path) as f:
            self.assertEqual(f.read(),
                             self.INPUT_FILESETS[fileset.basename])


class TestScanQualityLabelMatching(TestOnXnatMixin, BaseTestCase):

    # The labels are prefixed with the order that they should be uploaded
    # to XNAT with
    INPUT_FILESETS = {'1unusable': '1',
                      '2unlabelled': '1',
                      '3questionable': '1',
                      '4usable': '1'}

    def setUp(self):
        super(TestScanQualityLabelMatching, self).setUp()
        with self._connect() as login:
            xproject = login.projects[self.project]
            xsubject = login.classes.SubjectData(
                label='{}_{}'.format(self.project, self.SUBJECT),
                parent=xproject)
            xsession = login.classes.MrSessionData(
                label=self.session_label(),
                parent=xsubject)
            for scan in xsession.scans.values():
                if not scan.type.endswith('unlabelled'):
                    scan.quality = scan.type[1:]

    @unittest.skipIf(*SKIP_ARGS)
    def test_scan_label_quality(self):
        tmp_dir = tempfile.mkdtemp()
        repository = XnatRepo(
            project_id=self.project,
            server=SERVER, cache_dir=tmp_dir)
        tree = repository.tree(
            subject_ids=[self.SUBJECT],
            visit_ids=[self.VISIT])
        for accepted, expected in (
                (None, '1unusable'),
                ((None, 'questionable', 'usable'), '2unlabelled'),
                (('questionable', 'usable'), '3questionable'),
                ('usable', '4usable')):
            inpt = InputFilesets('dummy', order=0, valid_formats=text_format,
                                 acceptable_quality=accepted)
            matched = inpt.match(tree).item(subject_id=self.SUBJECT,
                                            visit_id=self.VISIT)
            self.assertEqual(matched.name, expected)



from __future__ import absolute_import
from past.builtins import basestring
import os
import tempfile
from arcana.utils import makedirs
import os.path as op # @IgnorePep8
import shutil
from arcana.utils import JSON_ENCODING
import stat
import time
import logging
import errno
import json
from zipfile import ZipFile, BadZipfile
from arcana.data import Fileset, Field
from arcana.repository.base import Repository
from arcana.exceptions import (
    ArcanaError, ArcanaUsageError, ArcanaFileFormatError,
    ArcanaWrongRepositoryError)
from arcana.pipeline.provenance import Record
from arcana.utils import dir_modtime, get_class_info, parse_value
import re
import xnat

logger = logging.getLogger('arcana')

special_char_re = re.compile(r'[^a-zA-Z_0-9]')
tag_parse_re = re.compile(r'\((\d+),(\d+)\)')

RELEVANT_DICOM_TAG_TYPES = set(('UI', 'CS', 'DA', 'TM', 'SH', 'LO',
                                'PN', 'ST', 'AS'))


class XnatRepo(Repository):
    """
    An 'Repository' class for XNAT repositories

    Parameters
    ----------
    server : str (URI)
        URI of XNAT server to connect to
    project_id : str
        The ID of the project in the XNAT repository
    cache_dir : str (path)
        Path to local directory to cache remote data in
    user : str
        Username with which to connect to XNAT with
    password : str
        Password to connect to the XNAT repository with
    check_md5 : bool
        Whether to check the MD5 digest of cached files before using. This
        checks for updates on the server since the file was cached
    race_cond_delay : int
        The amount of time to wait before checking that the required
        fileset has been downloaded to cache by another process has
        completed if they are attempting to download the same fileset
    session_filter : str
        A regular expression that is used to prefilter the discovered sessions
        to avoid having to retrieve metadata for them, and potentially speeding
        up the initialisation of the Study. Note that if the processing relies
        on summary derivatives (i.e. of 'per_visit/subject/study' frequency)
        then the filter should match all sessions in the Study's subject_ids
        and visit_ids.
    """

    type = 'xnat'

    SUMMARY_NAME = 'ALL'
    MD5_SUFFIX = '.__md5__.json'
    DERIVED_FROM_FIELD = '__derived_from__'
    PROV_SCAN = '__prov__'
    PROV_RESOURCE = 'PROV'
    depth = 2

    def __init__(self, server, project_id, cache_dir, user=None,
                 password=None, check_md5=True, race_cond_delay=30,
                 session_filter=None, **kwargs):
        super(XnatRepo, self).__init__(**kwargs)
        if not isinstance(server, basestring):
            raise ArcanaUsageError(
                "Invalid server url {}".format(server))
        self._project_id = project_id
        self._server = server
        self._cache_dir = cache_dir
        makedirs(self._cache_dir, exist_ok=True)
        self._user = user
        self._password = password
        self._race_cond_delay = race_cond_delay
        self._check_md5 = check_md5
        self._session_filter = session_filter
        self._login = None

    def __hash__(self):
        return (hash(self.server) ^
                hash(self.project_id) ^
                hash(self.cache_dir) ^
                hash(self._race_cond_delay) ^
                hash(self._check_md5))

    def __repr__(self):
        return ("{}(server={}, project_id={}, cache_dir={})"
                .format(type(self).__name__,
                        self.server, self.project_id,
                        self._cache_dir))

    def __eq__(self, other):
        try:
            return (self.server == other.server and
                    self._cache_dir == other._cache_dir and
                    self.project_id == other.project_id and
                    self.cache_dir == other.cache_dir and
                    self._race_cond_delay == other._race_cond_delay and
                    self._check_md5 == other._check_md5)
        except AttributeError:
            return False  # For comparison with other types

    @property
    def prov(self):
        return {
            'type': get_class_info(type(self)),
            'server': self.server,
            'project': self.project_id}

    @property
    def login(self):
        return self._login

    @property
    def project_id(self):
        return self._project_id

    @property
    def server(self):
        return self._server

    @property
    def cache_dir(self):
        return self._cache_dir

    @property
    def check_md5(self):
        return self._check_md5

    @property
    def session_filter(self):
        return (re.compile(self._session_filter)
                if self._session_filter is not None else None)

    def connect(self):
        """
        Parameters
        ----------
        prev_login : xnat.XNATSession
            An XNAT login that has been opened in the code that calls
            the method that calls login. It is wrapped in a
            NoExitWrapper so the returned connection can be used
            in a "with" statement in the method.
        """
        sess_kwargs = {}
        if self._user is not None:
            sess_kwargs['user'] = self._user
        if self._password is not None:
            sess_kwargs['password'] = self._password
        self._login = xnat.connect(server=self._server, **sess_kwargs)

    def disconnect(self):
        self._login.disconnect()
        self._login = None

    def get_fileset(self, fileset):
        """
        Caches a single fileset (if the 'path' attribute is accessed
        and it has not been previously cached for example

        Parameters
        ----------
        fileset : Fileset
            The fileset to cache
        prev_login : xnat.XNATSession
            An XNATSession object to use for the connection. A new
            one is created if one isn't provided

        Returns
        -------
        primary_path : str
            The path of the primary file once it has been cached
        aux_paths : dict[str, str]
            A dictionary containing a mapping of auxiliary file names to
            paths
        """
        if fileset.format is None:
            raise ArcanaUsageError(
                "Attempting to download {}, which has not been assigned a "
                "file format (see Fileset.formatted)".format(fileset))
        self._check_repository(fileset)
        with self:  # Connect to the XNAT repository if haven't already
            xsession = self.get_xsession(fileset)
            xscan = xsession.scans[fileset.name]
            # Set URI so we can retrieve checksums if required
            fileset.uri = xscan.uri
            fileset.id = xscan.id
            cache_path = self._cache_path(fileset)
            need_to_download = True
            if op.exists(cache_path):
                if self._check_md5:
                    md5_path = cache_path + XnatRepo.MD5_SUFFIX
                    try:
                        with open(md5_path, 'r') as f:
                            cached_checksums = json.load(f)
                        if cached_checksums == fileset.checksums:
                            need_to_download = False
                    except IOError:
                        pass
                else:
                    need_to_download = False
            if need_to_download:
                xresource = xscan.resources[fileset._resource_name]
                # The path to the directory which the files will be
                # downloaded to.
                tmp_dir = cache_path + '.download'
                try:
                    # Attempt to make tmp download directory. This will
                    # fail if another process (or previous attempt) has
                    # already created it. In that case this process will
                    # wait to see if that download finishes successfully,
                    # and if so use the cached version.
                    os.mkdir(tmp_dir)
                except OSError as e:
                    if e.errno == errno.EEXIST:
                        # Another process may be concurrently downloading
                        # the same file to the cache. Wait for
                        # 'race_cond_delay' seconds and then check that it
                        # has been completed or assume interrupted and
                        # redownload.
                        self._delayed_download(
                            tmp_dir, xresource, xscan, fileset,
                            xsession.label, cache_path,
                            delay=self._race_cond_delay)
                    else:
                        raise
                else:
                    self.download_fileset(
                        tmp_dir, xresource, xscan, fileset,
                        xsession.label, cache_path)
                    shutil.rmtree(tmp_dir)
        if not fileset.format.directory:
            (primary_path, aux_paths) = fileset.format.assort_files(
                op.join(cache_path, f) for f in os.listdir(cache_path))
        else:
            primary_path = cache_path
            aux_paths = None
        return primary_path, aux_paths

    def get_field(self, field):
        self._check_repository(field)
        with self:
            xsession = self.get_xsession(field)
            val = xsession.fields[field.name]
            val = val.replace('&quot;', '"')
            val = parse_value(val)
        return val

    def put_fileset(self, fileset):
        if fileset.format is None:
            raise ArcanaFileFormatError(
                "Format of {} needs to be set before it is uploaded to {}"
                .format(fileset, self))
        self._check_repository(fileset)
        # Open XNAT session
        with self:
            # Add session for derived scans if not present
            xsession = self.get_xsession(fileset)
            cache_path = self._cache_path(fileset)
            # Make session cache dir
            cache_path_dir = (op.dirname(cache_path)
                              if fileset.format.directory else cache_path)
            if os.path.exists(cache_path_dir):
                shutil.rmtree(cache_path_dir)
            os.makedirs(cache_path_dir, stat.S_IRWXU | stat.S_IRWXG)
            if fileset.format.directory:
                shutil.copytree(fileset.path, cache_path)
            else:
                # Copy primary file
                shutil.copyfile(fileset.path,
                                op.join(cache_path, fileset.fname))
                # Copy auxiliaries
                for sc_fname, sc_path in fileset.aux_file_fnames_and_paths:
                    shutil.copyfile(sc_path, op.join(cache_path, sc_fname))
            with open(cache_path + XnatRepo.MD5_SUFFIX, 'w',
                      **JSON_ENCODING) as f:
                json.dump(fileset.calculate_checksums(), f, indent=2)
            # Upload to XNAT
            xscan = self._login.classes.MrScanData(
                id=fileset.id, type=fileset.basename, parent=xsession)
            fileset.uri = xscan.uri
            # Select the first xnat_resource name to use to upload the data to
            resource_name = fileset.format.resource_names(self.type)[0]
            try:
                xresource = xscan.resources[resource_name]
            except KeyError:
                pass
            else:
                # Delete existing resource
                # TODO: probably should have check to see if we want to
                #       override it
                xresource.delete()
            xresource = xscan.create_resource(resource_name)
            if fileset.format.directory:
                for fname in os.listdir(fileset.path):
                    xresource.upload(op.join(fileset.path, fname), fname)
            else:
                xresource.upload(fileset.path, fileset.fname)
                for sc_fname, sc_path in fileset.aux_file_fnames_and_paths:
                    xresource.upload(sc_path, sc_fname)

    def put_field(self, field):
        self._check_repository(field)
        val = field.value
        if field.array:
            if field.dtype is str:
                val = ['"{}"'.format(v) for v in val]
            val = '[' + ','.join(str(v) for v in val) + ']'
        if field.dtype is str:
            val = '"{}"'.format(val)
        with self:
            xsession = self.get_xsession(field)
            xsession.fields[field.name] = val

    def put_record(self, record):
        base_cache_path = self._cache_path(record, name=self.PROV_SCAN)
        if not op.exists(base_cache_path):
            os.mkdir(base_cache_path)
        else:
            if not op.isdir(base_cache_path):
                raise ArcanaError(
                    "Base provenance cache path ('{}') should be a directory"
                    .format(base_cache_path))
        cache_path = op.join(base_cache_path, record.pipeline_name + '.json')
        record.save(cache_path)
        # TODO: Should also save digest of prov.json to check to see if it
        #       has been altered remotely
        xsession = self.get_xsession(record)
        xprov = self._login.classes.MrScanData(
            id=self.PROV_SCAN, type=self.PROV_SCAN, parent=xsession)
        # Delete existing provenance if present
        try:
            xresource = xprov.resources[record.pipeline_name]
        except KeyError:
            pass
        else:
            xresource.delete()
        # FIXME: should reuse the same resource for all provenance jsons
        xresource = xprov.create_resource(record.pipeline_name)
        xresource.upload(cache_path, op.basename(cache_path))

    def get_checksums(self, fileset):
        """
        Downloads the MD5 digests associated with the files in the file-set.
        These are saved with the downloaded files in the cache and used to
        check if the files have been updated on the server

        Parameters
        ----------
        resource : xnat.ResourceCatalog
            The xnat resource
        file_format : FileFormat
            The format of the fileset to get the checksums for. Used to
            determine the primary file within the resource and change the
            corresponding key in the checksums dictionary to '.' to match
            the way it is generated locally by Arcana.
        """
        if fileset.uri is None:
            raise ArcanaUsageError(
                "Can't retrieve checksums as URI has not been set for {}"
                .format(fileset))
        with self:
            checksums = {r['Name']: r['digest']
                         for r in self._login.get_json(fileset.uri + '/files')[
                             'ResultSet']['Result']}
        if not fileset.format.directory:
            # Replace the key corresponding to the primary file with '.' to
            # match the way that checksums are created by Arcana
            primary = fileset.format.assort_files(checksums.keys())[0]
            checksums['.'] = checksums.pop(primary)
        return checksums

    def find_data(self, subject_ids=None, visit_ids=None, **kwargs):
        """
        Find all filesets, fields and provenance records within an XNAT project

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned

        Returns
        -------
        filesets : list[Fileset]
            All the filesets found in the repository
        fields : list[Field]
            All the fields found in the repository
        records : list[Record]
            The provenance records found in the repository
        """
        subject_ids = self.convert_subject_ids(subject_ids)
        # Add derived visit IDs to list of visit ids to filter
        all_filesets = []
        all_fields = []
        all_records = []
        # Note we prefer the use of raw REST API calls here for performance
        # reasons over using XnatPy's data structures.
        with self:
            # Get map of internal subject IDs to subject labels in project
            subject_xids_to_labels = {
                s['ID']: s['label'] for s in self._login.get_json(
                    '/data/projects/{}/subjects'.format(self.project_id))[
                        'ResultSet']['Result']}
            # Get list of all sessions within project
            session_xids = [
                s['ID'] for s in self._login.get_json(
                    '/data/projects/{}/experiments'.format(self.project_id))[
                        'ResultSet']['Result']
                if (self.session_filter is None or
                    self.session_filter.match(s['label']))]
            for session_xid in session_xids:
                session_json = self._login.get_json(
                    '/data/projects/{}/experiments/{}'.format(
                        self.project_id, session_xid))['items'][0]
                subject_xid = session_json['data_fields']['subject_ID']
                subject_id = subject_xids_to_labels[subject_xid]
                session_label = session_json['data_fields']['label']
                session_uri = (
                    '/data/archive/projects/{}/subjects/{}/experiments/{}'
                    .format(self.project_id, subject_xid, session_xid))
                # Get field values. We do this first so we can check for the
                # DERIVED_FROM_FIELD to determine the correct session label and
                # study name
                field_values = {}
                try:
                    fields_json = next(
                        c['items'] for c in session_json['children']
                        if c['field'] == 'fields/field')
                except StopIteration:
                    pass
                else:
                    for js in fields_json:
                        try:
                            value = js['data_fields']['field']
                        except KeyError:
                            pass
                        else:
                            field_values[js['data_fields']['name']] = value
                # Extract study name and derived-from session
                if self.DERIVED_FROM_FIELD in field_values:
                    df_sess_label = field_values.pop(self.DERIVED_FROM_FIELD)
                    from_study = session_label[len(df_sess_label) + 1:]
                    session_label = df_sess_label
                else:
                    from_study = None
                # Strip subject ID from session label if required
                if session_label.startswith(subject_id + '_'):
                    visit_id = session_label[len(subject_id) + 1:]
                else:
                    visit_id = session_label
                # Strip project ID from subject ID if required
                if subject_id.startswith(self.project_id + '_'):
                    subject_id = subject_id[len(self.project_id) + 1:]
                # Check subject is summary or not and whether it is to be
                # filtered
                if subject_id == XnatRepo.SUMMARY_NAME:
                    subject_id = None
                elif not (subject_ids is None or subject_id in subject_ids):
                    continue
                # Check visit is summary or not and whether it is to be
                # filtered
                if visit_id == XnatRepo.SUMMARY_NAME:
                    visit_id = None
                elif not (visit_ids is None or visit_id in visit_ids):
                    continue
                # Determine frequency
                if (subject_id, visit_id) == (None, None):
                    frequency = 'per_study'
                elif visit_id is None:
                    frequency = 'per_subject'
                elif subject_id is None:
                    frequency = 'per_visit'
                else:
                    frequency = 'per_session'
                # Append fields
                for name, value in field_values.items():
                    value = value.replace('&quot;', '"')
                    all_fields.append(Field(
                        name=name, value=value, repository=self,
                        frequency=frequency,
                        subject_id=subject_id,
                        visit_id=visit_id,
                        from_study=from_study,
                        **kwargs))
                # Extract part of JSON relating to files
                try:
                    scans_json = next(
                        c['items'] for c in session_json['children']
                        if c['field'] == 'scans/scan')
                except StopIteration:
                    scans_json = []
                for scan_json in scans_json:
                    scan_id = scan_json['data_fields']['ID']
                    scan_type = scan_json['data_fields'].get('type', '')
                    scan_quality = scan_json['data_fields'].get('quality',
                                                                None)
                    scan_uri = '{}/scans/{}'.format(session_uri, scan_id)
                    try:
                        resources_json = next(
                            c['items'] for c in scan_json['children']
                            if c['field'] == 'file')
                    except StopIteration:
                        resources = {}
                    else:
                        resources = {js['data_fields']['label']:
                                     js['data_fields'].get('format', None)
                                     for js in resources_json}
                    # Remove auto-generated snapshots directory
                    resources.pop('SNAPSHOTS', None)
                    if scan_type == self.PROV_SCAN:
                        # Download provenance JSON files and parse into
                        # records
                        temp_dir = tempfile.mkdtemp()
                        try:
                            with tempfile.TemporaryFile() as temp_zip:
                                self._login.download_stream(
                                    scan_uri + '/files', temp_zip,
                                    format='zip')
                                with ZipFile(temp_zip) as zip_file:
                                    zip_file.extractall(temp_dir)
                            for base_dir, _, fnames in os.walk(temp_dir):
                                for fname in fnames:
                                    if fname.endswith('.json'):
                                        pipeline_name = fname[:-len('.json')]
                                        json_path = op.join(base_dir, fname)
                                        all_records.append(
                                            Record.load(
                                                pipeline_name, frequency,
                                                subject_id, visit_id,
                                                from_study, json_path))
                        finally:
                            shutil.rmtree(temp_dir, ignore_errors=True)
                    else:
                        for resource in resources:
                            all_filesets.append(Fileset(
                                scan_type, id=scan_id, uri=scan_uri,
                                repository=self, frequency=frequency,
                                subject_id=subject_id, visit_id=visit_id,
                                from_study=from_study, quality=scan_quality,
                                resource_name=resource, **kwargs))
                logger.debug("Found node {}:{} on {}:{}".format(
                    subject_id, visit_id, self.server, self.project_id))
        return all_filesets, all_fields, all_records

    def convert_subject_ids(self, subject_ids):
        """
        Convert subject ids to strings if they are integers
        """
        # TODO: need to make this generalisable via a
        #       splitting+mapping function passed to the repository
        if subject_ids is not None:
            subject_ids = set(
                ('{:03d}'.format(s)
                 if isinstance(s, int) else s) for s in subject_ids)
        return subject_ids

    def extract_subject_id(self, xsubject_label):
        """
        This assumes that the subject ID is prepended with
        the project ID.
        """
        return xsubject_label[(len(self.project_id) + 1):]

    def extract_visit_id(self, xsession_label):
        """
        This assumes that the session ID is preprended
        """
        return '_'.join(xsession_label.split('_')[2:])

    def dicom_header(self, fileset):
        def convert(val, code):  # @IgnorePep8
            if code == 'TM':
                try:
                    val = float(val)
                except ValueError:
                    pass
            elif code == 'CS':
                val = val.split('\\')
            return val
        with self:
            response = self._login.get(
                '/REST/services/dicomdump?src=' +
                fileset.uri[len('/data'):]).json()['ResultSet']['Result']
        hdr = {tag_parse_re.match(t['tag1']).groups(): convert(t['value'],
                                                               t['vr'])
               for t in response if (tag_parse_re.match(t['tag1']) and
                                     t['vr'] in RELEVANT_DICOM_TAG_TYPES)}
        return hdr

    def download_fileset(self, tmp_dir, xresource, xscan, fileset,
                          session_label, cache_path):
        # Download resource to zip file
        zip_path = op.join(tmp_dir, 'download.zip')
        with open(zip_path, 'wb') as f:
            xresource.xnat_session.download_stream(
                xresource.uri + '/files', f, format='zip', verbose=True)
        checksums = self.get_checksums(fileset)
        # Extract downloaded zip file
        expanded_dir = op.join(tmp_dir, 'expanded')
        try:
            with ZipFile(zip_path) as zip_file:
                zip_file.extractall(expanded_dir)
        except BadZipfile as e:
            raise ArcanaError(
                "Could not unzip file '{}' ({})"
                .format(xresource.id, e))
        data_path = op.join(
            expanded_dir, session_label, 'scans',
            (xscan.id + '-' + special_char_re.sub('_', xscan.type)),
            'resources', xresource.label, 'files')
        # Remove existing cache if present
        try:
            shutil.rmtree(cache_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise e
        shutil.move(data_path, cache_path)
        with open(cache_path + XnatRepo.MD5_SUFFIX, 'w',
                  **JSON_ENCODING) as f:
            json.dump(checksums, f, indent=2)

    def _delayed_download(self, tmp_dir, xresource, xscan, fileset,
                          session_label, cache_path, delay):
        logger.info("Waiting {} seconds for incomplete download of '{}' "
                    "initiated another process to finish"
                    .format(delay, cache_path))
        initial_mod_time = dir_modtime(tmp_dir)
        time.sleep(delay)
        if op.exists(cache_path):
            logger.info("The download of '{}' has completed "
                        "successfully in the other process, continuing"
                        .format(cache_path))
            return
        elif initial_mod_time != dir_modtime(tmp_dir):
            logger.info(
                "The download of '{}' hasn't completed yet, but it has"
                " been updated.  Waiting another {} seconds before "
                "checking again.".format(cache_path, delay))
            self._delayed_download(tmp_dir, xresource, xscan,
                                   fileset,
                                   session_label, cache_path, delay)
        else:
            logger.warning(
                "The download of '{}' hasn't updated in {} "
                "seconds, assuming that it was interrupted and "
                "restarting download".format(cache_path, delay))
            shutil.rmtree(tmp_dir)
            os.mkdir(tmp_dir)
            self.download_fileset(
                tmp_dir, xresource, xscan, fileset, session_label,
                cache_path)

    def get_xsession(self, item):
        """
        Returns the XNAT session and cache dir corresponding to the
        item.
        """
        subj_label, sess_label = self._get_item_labels(item)
        with self:
            xproject = self._login.projects[self.project_id]
            try:
                xsubject = xproject.subjects[subj_label]
            except KeyError:
                xsubject = self._login.classes.SubjectData(
                    label=subj_label, parent=xproject)
            try:
                xsession = xsubject.experiments[sess_label]
            except KeyError:
                xsession = self._login.classes.MrSessionData(
                    label=sess_label, parent=xsubject)
                if item.derived:
                    xsession.fields[
                        self.DERIVED_FROM_FIELD] = self._get_item_labels(
                            item, no_from_study=True)[1]
        return xsession

    def _get_item_labels(self, item, no_from_study=False):
        """
        Returns the labels for the XNAT subject and sessions given
        the frequency and provided IDs.
        """
        subject_id = self.inv_map_subject_id(item.subject_id)
        visit_id = self.inv_map_visit_id(item.visit_id)
        subj_label, sess_label = self._get_labels(
            item.frequency, subject_id, visit_id)
        if not no_from_study and item.from_study is not None:
            sess_label += '_' + item.from_study
        return (subj_label, sess_label)

    def _get_labels(self, frequency, subject_id=None, visit_id=None):
        """
        Returns the labels for the XNAT subject and sessions given
        the frequency and provided IDs.
        """
        if frequency == 'per_session':
            subj_label = '{}_{}'.format(self.project_id,
                                        subject_id)
            sess_label = '{}_{}_{}'.format(self.project_id,
                                           subject_id,
                                           visit_id)
        elif frequency == 'per_subject':
            subj_label = '{}_{}'.format(self.project_id,
                                        subject_id)
            sess_label = '{}_{}_{}'.format(self.project_id,
                                           subject_id,
                                           self.SUMMARY_NAME)
        elif frequency == 'per_visit':
            subj_label = '{}_{}'.format(self.project_id,
                                        self.SUMMARY_NAME)
            sess_label = '{}_{}_{}'.format(self.project_id,
                                           self.SUMMARY_NAME,
                                           visit_id)
        elif frequency == 'per_study':
            subj_label = '{}_{}'.format(self.project_id,
                                        self.SUMMARY_NAME)
            sess_label = '{}_{}_{}'.format(self.project_id,
                                           self.SUMMARY_NAME,
                                           self.SUMMARY_NAME)
        else:
            assert False
        return (subj_label, sess_label)

    def _cache_path(self, fileset, name=None):
        subj_dir, sess_dir = self._get_item_labels(fileset)
        cache_dir = op.join(self._cache_dir, self.project_id,
                            subj_dir, sess_dir)
        makedirs(cache_dir, exist_ok=True)
        return op.join(cache_dir, fileset.name if name is None else name)

    def _check_repository(self, item):
        if item.repository is not self:
            raise ArcanaWrongRepositoryError(
                "{} is from {} instead of {}".format(item, item.repository,
                                                     self))



from __future__ import print_function
from past.builtins import basestring
import os.path as op
import os
from arcana.utils import makedirs
import subprocess as sp
import shutil
from unittest import TestCase
import errno
import sys
import json
from arcana.utils import JSON_ENCODING
import filecmp
from copy import deepcopy
import logging
import arcana
from arcana.data import Fileset
from arcana.utils import classproperty
from arcana.repository.basic import BasicRepo
from arcana.processor import SingleProc
from arcana.environment import StaticEnv
from arcana.exceptions import ArcanaError
from arcana.exceptions import ArcanaUsageError

logger = logging.getLogger('arcana')
logger.setLevel(logging.WARNING)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

wf_logger = logging.getLogger('nipype.workflow')
wf_logger.setLevel(logging.WARNING)
intf_logger = logging.getLogger('nipype.interface')
intf_logger.setLevel(logging.WARNING)

logging.getLogger("urllib3").setLevel(logging.WARNING)


class BaseTestCase(TestCase):

    SUBJECT = 'SUBJECT'
    VISIT = 'VISIT'
    SESSION = (SUBJECT, VISIT)

    # Whether to copy reference filesets from reference directory
    INPUTS_FROM_REF_DIR = False

    # The path to the test directory, which should sit along side the
    # the package directory. Note this will not work when Arcana
    # is installed by a package manager.
    BASE_TEST_DIR = op.abspath(op.join(
        op.dirname(arcana.__file__), '..', 'test'))

    @classproperty
    @classmethod
    def test_data_dir(cls):
        try:
            return cls._test_data_dir
        except AttributeError:
            try:
                cls._test_data_dir = os.environ['ARCANA_TEST_DATA']
                makedirs(cls._test_data_dir, exist_ok=True)
            except KeyError:
                cls._test_data_dir = op.join(cls.BASE_TEST_DIR, 'data')
            return cls._test_data_dir

    @classproperty
    @classmethod
    def unittest_root(cls):
        return op.join(cls.BASE_TEST_DIR, 'unittests')

    @classproperty
    @classmethod
    def repository_path(cls):
        return op.join(cls.test_data_dir, 'repository')

    @classproperty
    @classmethod
    def work_path(cls):
        return op.join(cls.test_data_dir, 'work')

    @classproperty
    @classmethod
    def cache_path(cls):
        return op.join(cls.test_data_dir, 'cache')

    @classproperty
    @classmethod
    def ref_path(cls):
        return op.join(cls.BASE_TEST_DIR, 'data', 'reference')

    def setUp(self):
        self.reset_dirs()
        if self.INPUTS_FROM_REF_DIR:
            filesets = {}
            # Unzip reference directory if required
            if not os.path.exists(self.ref_dir) and os.path.exists(
                    self.ref_dir + '.tar.gz'):
                sp.check_call(
                    'tar xzf {}.tar.gz'.format(self.ref_dir),
                    shell=True, cwd=os.path.dirname(self.ref_dir))
            for fname in os.listdir(self.ref_dir):
                if fname.startswith('.'):
                    continue
                fileset = Fileset.from_path(op.join(self.ref_dir,
                                                    fname))
                fileset.format = fileset.detect_format(self.REF_FORMATS)
                filesets[fileset.name] = fileset
        else:
            filesets = getattr(self, 'INPUT_FILESETS', None)
        self.add_session(filesets=filesets,
                         fields=getattr(self, 'INPUT_FIELDS', None))

    def add_session(self, filesets=None, fields=None, project_dir=None,
                    subject=None, visit=None):
        if project_dir is None:
            project_dir = self.project_dir
        if filesets is None:
            filesets = {}
        if subject is None:
            subject = self.SUBJECT
        if visit is None:
            visit = self.VISIT
        session_dir = op.join(project_dir, subject, visit)
        os.makedirs(session_dir)
        for name, fileset in list(filesets.items()):
            if isinstance(fileset, Fileset):
                if fileset.format is None:
                    raise ArcanaError(
                        "Need to provide format for fileset to add to test "
                        "dataset ({}) in {}".format(fileset, self))
                dst_path = op.join(session_dir,
                                   name + fileset.format.ext_str)
                if fileset.format.directory:
                    shutil.copytree(fileset.path, dst_path)
                else:
                    shutil.copy(fileset.path, dst_path)
            elif isinstance(fileset, basestring):
                # Write string as text file
                with open(op.join(session_dir,
                                  name + '.txt'), 'w') as f:
                    f.write(fileset)
            else:
                raise ArcanaError(
                    "Unrecognised fileset ({}) in {} test setup. Can "
                    "be either a Fileset or basestring object"
                    .format(fileset, self))
        if fields is not None:
            with open(op.join(session_dir,
                              BasicRepo.FIELDS_FNAME), 'w',
                      **JSON_ENCODING) as f:
                json.dump(fields, f, indent=2)

    def delete_project(self, project_dir):
        # Clean out any existing repository files
        shutil.rmtree(project_dir, ignore_errors=True)

    def reset_dirs(self):
        shutil.rmtree(self.project_dir, ignore_errors=True)
        shutil.rmtree(self.work_dir, ignore_errors=True)
        self.create_dirs()

    def create_dirs(self):
        for d in (self.project_dir, self.work_dir):
            if not op.exists(d):
                os.makedirs(d)

    @property
    def repository_tree(self):
        return self.repository.tree()

    @property
    def xnat_session_name(self):
        return '{}_{}'.format(self.XNAT_TEST_PROJECT, self.name)

    @property
    def session_dir(self):
        return self.get_session_dir(self.SUBJECT, self.VISIT)

    def derived_session_dir(self, from_study=None):
        if from_study is None:
            from_study = self.STUDY_NAME
        return op.join(self.session_dir, from_study)

    @property
    def session(self):
        return self.repository_tree.subject(
            self.SUBJECT).session(self.VISIT)

    @property
    def filesets(self):
        return self.session.filesets

    @property
    def fields(self):
        return self.session.fields

    @property
    def repository(self):
        return self.local_repository

    @property
    def local_repository(self):
        try:
            return self._local_repository
        except AttributeError:
            self._local_repository = BasicRepo(self.project_dir,
                                                         depth=2)
            return self._local_repository

    @property
    def processor(self):
        return SingleProc(self.work_dir)

    @property
    def environment(self):
        return StaticEnv()

    @property
    def project_dir(self):
        return op.join(self.repository_path, self.name)

    @property
    def work_dir(self):
        return op.join(self.work_path, self.name)

    @property
    def cache_dir(self):
        return op.join(self.cache_path, self.name)

    @property
    def ref_dir(self):
        return op.join(self.ref_path, self.name)

    @property
    def name(self):  # @NoSelf
        return self._get_name(type(self))

    @property
    def project_id(self):
        return self.name  # To allow override in deriving classes

    @classmethod
    def _get_name(cls, name_cls=None, sep='_'):
        """
        Get unique name for test class from module path and its class name to
        be used for storing test data on XNAT and creating unique work/project
        dirs
        """
        if name_cls is None:
            name_cls = cls
        module_path = op.abspath(sys.modules[name_cls.__module__].__file__)
        rel_module_path = module_path[(len(name_cls.unittest_root) + 1):]
        path_parts = rel_module_path.split(op.sep)
        module_name = (''.join(path_parts[:-1]) +
                       op.splitext(path_parts[-1])[0][5:]).upper()
        test_class_name = name_cls.__name__.upper()
        if test_class_name.startswith('TEST'):
            test_class_name = test_class_name[4:]
        return module_name + sep + test_class_name

    def create_study(self, study_cls, name, inputs, repository=None,
                     processor=None, environment=None, **kwargs):
        """
        Creates a study using default repository and processors.

        Parameters
        ----------
        study_cls : Study
            The class to initialise
        name : str
            Name of the study
        inputs : List[BaseSpec]
            List of inputs to the study
        repository : Repository | None
            The repository to use (a default local repository is used if one
            isn't provided
        processor : Processor | None
            The processor to use (a default SingleProc is used if one
            isn't provided
        """
        if repository is None:
            repository = self.repository
        if processor is None:
            processor = self.processor
        if environment is None:
            environment = self.environment
        return study_cls(
            name=name,
            repository=repository,
            processor=processor,
            environment=environment,
            inputs=inputs,
            **kwargs)

    def assertFilesetCreated(self, fileset):
        self.assertTrue(
            op.exists(fileset.path),
            ("{} was not created by unittest".format(fileset)))

    def assertContentsEqual(self, collection, reference, context=None):
        if isinstance(collection, Fileset):
            collection = [collection]
        if isinstance(reference, (basestring, int, float)):
            if len(collection) != 1:
                raise ArcanaUsageError(
                    "Multi-subject/visit collections cannot be compared"
                    " against a single contents string (list or dict "
                    "should be provided)")
            references = [str(reference)]
            filesets = list(collection)
        elif isinstance(reference, dict):
            references = []
            filesets = []
            for subj_id, subj_dct in references.items():
                for visit_id, ref_value in subj_dct.items():
                    references.append(str(ref_value))
                    filesets.append(collection.item(subject_id=subj_id,
                                                    visit_id=visit_id))
        elif isinstance(reference, (list, tuple)):
            references = [str(r) for r in reference]
            filesets = list(collection)
            if len(references) != len(filesets):
                raise ArcanaUsageError(
                    "Number of provided references ({}) does not match"
                    " size of collection ({})".format(len(references),
                                                      len(filesets)))
        else:
            raise ArcanaUsageError(
                "Unrecognised format for reference ({})"
                .format(reference))
        for fileset, ref in zip(filesets, references):
            with open(fileset.path) as f:
                contents = f.read()
            msg = ("Contents of {} ({}) do not match reference ({})"
                   .format(fileset, contents, ref))
            if context is not None:
                msg += 'for ' + context
            self.assertEqual(contents, ref, msg)

    def assertCreated(self, fileset):
        self.assertTrue(
            os.path.exists(fileset.path),
            "{} was not created".format(fileset))

    def assertField(self, name, ref_value, from_study, subject=None,
                    visit=None, frequency='per_session',
                    to_places=None):
        esc_name = from_study + '_' + name
        output_dir = self.get_session_dir(subject, visit, frequency)
        try:
            with open(op.join(output_dir,
                              BasicRepo.FIELDS_FNAME)) as f:
                fields = json.load(f, 'rb')
        except IOError as e:
            if e.errno == errno.ENOENT:
                raise ArcanaError(
                    "No fields were created by pipeline in study '{}'"
                    .format(from_study))
        try:
            value = fields[esc_name]
        except KeyError:
            raise ArcanaError(
                "Field '{}' was not created by pipeline in study '{}'. "
                "Created fields were ('{}')"
                .format(esc_name, from_study, "', '".join(fields)))
        msg = ("Field value '{}' for study '{}', {}, does not match "
               "reference value ({})".format(name, from_study, value,
                                             ref_value))
        if to_places is not None:
            self.assertAlmostEqual(
                value, ref_value, to_places,
                '{} to {} decimal places'.format(msg, to_places))
        else:
            self.assertEqual(value, ref_value, msg)

    def assertFilesetsEqual(self, fileset1, fileset2, error_msg=None):
        msg = "{} does not match {}".format(fileset1, fileset2)
        if msg is not None:
            msg += ':\n' + error_msg
        self.assertTrue(filecmp.cmp(fileset1.path, fileset2.path,
                                    shallow=False), msg=msg)

    def assertStatEqual(self, stat, fileset_name, target, from_study,
                        subject=None, visit=None,
                        frequency='per_session'):
            val = float(sp.check_output(
                'mrstats {} -output {}'.format(
                    self.output_file_path(
                        fileset_name, from_study,
                        subject=subject, visit=visit,
                        frequency=frequency),
                    stat),
                shell=True))
            self.assertEqual(
                val, target, (
                    "{} value of '{}' ({}) does not equal target ({}) "
                    "for subject {} visit {}"
                    .format(stat, fileset_name, val, target,
                            subject, visit)))

    def assertImagesAlmostMatch(self, out, ref, mean_threshold,
                                stdev_threshold, from_study):
        out_path = self.output_file_path(out, from_study)
        ref_path = self.ref_file_path(ref)
        # Should probably look into ITK fuzzy matching methods
        cmd = ("mrcalc -quiet {a} {b} -subtract - | mrstats - | "
               "grep -v channel | awk '{{print $4 \" \" $6}}'"
               .format(a=out_path, b=ref_path))
        out = sp.check_output(cmd, shell=True)
        mean, stdev = (float(x) for x in out.split())
        self.assert_(
            abs(mean) < mean_threshold and stdev < stdev_threshold,
            ("Mean ({mean}) or standard deviation ({stdev}) of difference "
             "between images {a} and {b} differ more than threshold(s) "
             "({thresh_mean} and {thresh_stdev} respectively)"
             .format(mean=mean, stdev=stdev, thresh_mean=mean_threshold,
                     thresh_stdev=stdev_threshold, a=out_path, b=ref_path)))

    def get_session_dir(self, subject=None, visit=None,
                        frequency='per_session', from_study=None):
        if subject is None and frequency in ('per_session', 'per_subject'):
            subject = self.SUBJECT
        if visit is None and frequency in ('per_session', 'per_visit'):
            visit = self.VISIT
        if frequency == 'per_session':
            assert subject is not None
            assert visit is not None
            path = op.join(self.project_dir, subject, visit)
        elif frequency == 'per_subject':
            assert subject is not None
            assert visit is None
            path = op.join(
                self.project_dir, subject,
                BasicRepo.SUMMARY_NAME)
        elif frequency == 'per_visit':
            assert visit is not None
            assert subject is None
            path = op.join(self.project_dir,
                           BasicRepo.SUMMARY_NAME, visit)
        elif frequency == 'per_study':
            assert subject is None
            assert visit is None
            path = op.join(self.project_dir,
                           BasicRepo.SUMMARY_NAME,
                           BasicRepo.SUMMARY_NAME)
        else:
            assert False
        if from_study is not None:
            path = op.join(path, from_study)
        return op.abspath(path)

    @classmethod
    def remove_generated_files(cls, study=None):
        # Remove derived filesets
        for fname in os.listdir(cls.get_session_dir()):
            if study is None or fname.startswith(study + '_'):
                os.remove(op.join(cls.get_session_dir(), fname))

    def output_file_path(self, fname, from_study, subject=None, visit=None,
                         frequency='per_session', **kwargs):
        return op.join(
            self.get_session_dir(subject=subject, visit=visit,
                                 frequency=frequency,
                                 from_study=from_study, **kwargs),
            fname)

    def ref_file_path(self, fname, subject=None, session=None):
        return op.join(self.session_dir, fname,
                            subject=subject, session=session)


class BaseMultiSubjectTestCase(BaseTestCase):

    SUMMARY_NAME = BasicRepo.SUMMARY_NAME

    def setUp(self):
        self.reset_dirs()
        self.add_sessions()

    def add_sessions(self):
        self.local_tree = deepcopy(self.input_tree)
        for node in self.local_tree:
            for fileset in node.filesets:
                fileset._repository = self.local_repository
                fileset._path = op.join(
                    fileset.repository.fileset_path(fileset))
                session_path = op.dirname(fileset.path)
                self._make_dir(session_path)
                contents = self.DATASET_CONTENTS[fileset.name]
                if fileset.format.aux_files:
                    fileset._aux_files = {}
                    for (aux_name,
                         aux_path) in fileset.format.default_aux_file_paths(
                            fileset._path).items():
                        fileset._aux_files[aux_name] = aux_path
                        with open(aux_path, 'w') as f:
                            f.write(str(contents[aux_name]))
                    contents = contents['.']
                with open(fileset.path, 'w') as f:
                    f.write(str(contents))
                if fileset.derived:
                    self._make_dir(op.join(session_path,
                                           BasicRepo.PROV_DIR))
            for field in node.fields:
                fields_path = self.local_repository.fields_json_path(field)
                if op.exists(fields_path):
                    with open(fields_path, **JSON_ENCODING) as f:
                        dct = json.load(f)
                else:
                    if not op.exists(op.dirname(fields_path)):
                        os.makedirs(op.dirname(fields_path))
                    dct = {}
                dct[field.name] = field.value
                with open(fields_path, 'w', **JSON_ENCODING) as f:
                    json.dump(dct, f, indent=2)
                if field.derived:
                    self._make_dir(op.join(session_path,
                                           BasicRepo.PROV_DIR))

    @property
    def subject_ids(self):
        return (d for d in os.listdir(self.project_dir)
                if d != self.SUMMARY_NAME)

    def visit_ids(self, subject_id):
        subject_dir = op.join(self.project_dir, subject_id)
        return (d for d in os.listdir(subject_dir)
                if d != self.SUMMARY_NAME)

    def session_dir(self, subject, visit):
        return self.get_session_dir(subject, visit)

    def get_session_dir(self, subject, visit, **kwargs):
        return super(BaseMultiSubjectTestCase, self).get_session_dir(
            subject=subject, visit=visit, **kwargs)

    def _make_dir(self, path):
        makedirs(path, exist_ok=True)
        return path


class DummyTestCase(BaseTestCase):

    def __init__(self):
        self.setUp()

    def __del__(self):
        self.tearDown()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def assert_(self, statement, message=None):
        if not statement:
            message = "'{}' is not true".format(statement)
            print(message)
        else:
            print("Test successful")

    def assertEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print(message)
        else:
            print("Test successful")

    def assertAlmostEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print(message)
        else:
            print("Test successful")

    def assertLess(self, first, second, message=None):
        if first >= second:
            if message is None:
                message = '{} is not less than {}'.format(repr(first),
                                                          repr(second))
            print(message)
        else:
            print("Test successful")



from builtins import object
import os
import os.path as op
from collections import defaultdict
from arcana.exceptions import (
    ArcanaUsageError, ArcanaNoConverterError, ArcanaFileFormatError)
from nipype.interfaces.utility import IdentityInterface
from arcana.utils.interfaces import (
    ZipDir, UnzipDir, TarGzDir, UnTarGzDir)
from arcana.utils import split_extension
import logging


logger = logging.getLogger('arcana')


class FileFormat(object):
    """
    Defines a format for a fileset (e.g. DICOM, NIfTI, Matlab file)

    Parameters
    ----------
    name : str
        A name for the data format
    extension : str
        The extension of the format
    desc : str
        A description of what the format is and ideally a link to its
        documentation
    directory : bool
        Whether the format is a directory or a file
    within_dir_exts : List[str]
        A list of extensions that are found within the top level of
        the directory (for directory formats). Used to identify
        formats from paths.
    aux_files : dict[str, str]
        A dictionary of side cars (e.g. header or NIfTI json side cars) aside
        from the primary file, along with their expected extension.
        Automatically they will be assumed to be located adjancent to the
        primary file, with the same base name and this extension. However, in
        the initialisation of the fileset, alternate locations can be specified
    resource_names : Dict[str, List[str]]
        A dictionary mapping the name of a repository type to a list of
        alternate names to use to load the file format with (when the format is
        saved by format name, e.g. XNAT, instead of a file with an extension)
    """

    def __init__(self, name, extension=None, desc='',
                 directory=False, within_dir_exts=None,
                 aux_files=None, resource_names=None):
        if not name.islower():
            raise ArcanaUsageError(
                "All data format names must be lower case ('{}')"
                .format(name))
        if extension is None and not directory:
            raise ArcanaUsageError(
                "Extension for '{}' format can only be None if it is a "
                "directory".format(name))
        self._name = name
        self._extension = extension
        self._desc = desc
        self._directory = directory
        if within_dir_exts is not None:
            if not directory:
                raise ArcanaUsageError(
                    "'within_dir_exts' keyword arg is only valid "
                    "for directory data formats, not '{}'".format(name))
            within_dir_exts = frozenset(within_dir_exts)
        self._within_dir_exts = within_dir_exts
        self._converters = {}
        self._resource_names = (resource_names
                                if resource_names is not None else {})
        self._aux_files = aux_files if aux_files is not None else {}
        for sc_name, sc_ext in self.aux_files.items():
            if sc_ext == self.ext:
                raise ArcanaUsageError(
                    "Extension for side car '{}' cannot be the same as the "
                    "primary file ('{}')".format(sc_name, sc_ext))

    def __eq__(self, other):
        try:
            return (
                self._name == other._name and
                self._extension == other._extension and
                self._desc == other._desc and
                self._directory == other._directory and
                self._within_dir_exts ==
                other._within_dir_exts and
                self._resource_names == other._resource_names and
                self.aux_files == other.aux_files)
        except AttributeError:
            return False

    def __hash__(self):
        return (
            hash(self._name) ^
            hash(self._extension) ^
            hash(self._desc) ^
            hash(self._directory) ^
            hash(self._within_dir_exts) ^
            hash(tuple((repo_type, tuple(self._resource_names[repo_type]))
                       for repo_type in sorted(self._resource_names))) ^
            hash(tuple(sorted(self.aux_files.items()))))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return ("FileFormat(name='{}', extension='{}', directory={}{})"
                .format(self.name, self.extension, self.directory,
                        (', within_dir_extension={}'.format(
                            self.within_dir_exts)
                         if self.directory else '')))

    def __str__(self):
        return self.name

    @property
    def name(self):
        return self._name

    @property
    def extension(self):
        return self._extension

    @property
    def extensions(self):
        return tuple([self._extension] + sorted(self.aux_file_exts))

    @property
    def ext(self):
        return self._extension

    @property
    def ext_str(self):
        return self.extension if self.extension is not None else ''

    @property
    def desc(self):
        return self._desc

    @property
    def directory(self):
        return self._directory

    @property
    def aux_files(self):
        return self._aux_files

    def resource_names(self, repo_type):
        """
        Names of resources used to store the format on a given repository type.
        Defaults to the name of the name of the format
        """
        try:
            names = self._resource_names[repo_type]
        except KeyError:
            names = [self.name, self.name.upper()]
        return names

    def default_aux_file_paths(self, primary_path):
        """
        Get the default paths for auxiliary files relative to the path of the
        primary file, i.e. the same name as the primary path with a different
        extension

        Parameters
        ----------
        primary_path : str
            Path to the primary file in the fileset

        Returns
        -------
        aux_paths : dict[str, str]
            A dictionary of auxiliary file names and default paths
        """
        return dict((n, primary_path[:-len(self.ext)] + ext)
                    for n, ext in self.aux_files.items())

    @property
    def aux_file_exts(self):
        return frozenset(self._aux_files.values())

    @property
    def within_dir_exts(self):
        return self._within_dir_exts

    def converter_from(self, file_format, **kwargs):
        if file_format == self:
            return IdentityConverter(file_format, self)
        try:
            matching_format, converter_cls = self._converters[file_format.name]
        except KeyError:
            raise ArcanaNoConverterError(
                "There is no converter to convert {} to {} (available: {})"
                .format(file_format, self,
                        ', '.join(
                            '{} <- {}'.format(k, v)
                            for k, v in self._converters.items())))
        if file_format != matching_format:
            raise ArcanaNoConverterError(
                "{} matches the name of a format that {} can be converted from"
                " but is not the identical".format(file_format,
                                                   matching_format))
        return converter_cls(file_format, self, **kwargs)

    @property
    def convertable_from(self):
        """
        A list of formats that the current format can be converted from
        """
        return (f for f, _ in self._converters.values())

    def assort_files(self, candidates):
        """
        Assorts candidate files into primary and auxiliary (and ignored) files
        corresponding to the format by their file extensions. Can be overridden
        in specialised subclasses to assort files based on other
        characteristics

        Parameters
        ----------
        candidates : list[str]
            The list of filenames to assort

        Returns
        -------
        primary_file : str
            Path to the selected primary file
        aux_files : dict[str, str]
            A dictionary mapping the auxiliary file name to the selected path
        """
        by_ext = defaultdict(list)
        for path in candidates:
            by_ext[split_extension(path)[1].lower()].append(path)
        try:
            primary_file = by_ext[self.ext]
        except KeyError:
            raise ArcanaFileFormatError(
                "No files match primary file extension of {} out of "
                "potential candidates of {}"
                .format(self, "', '".join(candidates)))
        if not primary_file:
            raise ArcanaFileFormatError(
                "No potential files for primary file of {}".format(self))
        elif len(primary_file) > 1:
            raise ArcanaFileFormatError(
                "Multiple potential files for '{}' primary file of {}"
                .format("', '".join(primary_file), self))
        else:
            primary_file = primary_file[0]
        aux_files = {}
        for aux_name, aux_ext in self.aux_files.items():
            aux = by_ext[aux_ext]
            if not aux:
                raise ArcanaFileFormatError(
                    "No files match auxiliary file extension '{}' of {} out of"
                    " potential candidates of {}"
                    .format(aux_ext, self, "', '".join(candidates)))
            elif len(aux) > 1:
                raise ArcanaFileFormatError(
                    "Multiple potential files for '{}' auxiliary file ext. "
                    "({}) of {}".format("', '".join(aux),
                                        self))
            else:
                aux_files[aux_name] = aux[0]
        return primary_file, aux_files

    def matches(self, fileset):
        """
        Checks to see whether the format matches the given fileset

        Parameters
        ----------
        fileset : Fileset
            The fileset to check
        """
        if fileset._resource_name is not None:
            return (fileset._resource_name in self.resource_names(
                fileset.repository.type))
        elif self.directory:
            if op.isdir(fileset.path):
                if self.within_dir_exts is None:
                    return True
                else:
                    # Get set of all extensions in the directory
                    return self.within_dir_exts == frozenset(
                        split_extension(f)[1] for f in os.listdir(fileset.path)
                        if not f.startswith('.'))
            else:
                return False
        else:
            if op.isfile(fileset.path):
                all_paths = [fileset.path]
                if fileset._potential_aux_files is not None:
                    all_paths += fileset._potential_aux_files
                try:
                    primary_path = self.assort_files(all_paths)[0]
                except ArcanaFileFormatError:
                    return False
                else:
                    return primary_path == fileset.path
            else:
                return False

    def set_converter(self, file_format, converter):
        """
        Register a Converter and the FileFormat that it is able to convert from

        Parameters
        ----------
        converter : Converter
            The converter to register
        file_format : FileFormat
            The file format that can be converted into this format
        """
        self._converters[file_format.name] = (file_format, converter)


class Converter(object):
    """
    Base class for all Arcana data format converters

    Parameters
    ----------
    input_format : FileFormat
        The input format to convert from
    output_format : FileFormat
        The output format to convert to
    """

    requirements = []

    def __init__(self, input_format, output_format, wall_time=None,
                 mem_gb=None):
        self._input_format = input_format
        self._output_format = output_format
        self._wall_time = wall_time
        self._mem_gb = mem_gb

    def __eq__(self, other):
        return (self.input_format == self.input_format and
                self._output_format == other.output_format)

    @property
    def input_format(self):
        return self._input_format

    @property
    def output_format(self):
        return self._output_format

    @property
    def interface(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def input(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def output(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def mem_gb(self):
        return self._mem_gb

    @property
    def wall_time(self):
        return self._wall_time

    def __repr__(self):
        return "{}(input_format={}, output_format={})".format(
            type(self).__name__, self.input_format, self.output_format)


class IdentityConverter(Converter):

    requirements = []
    interface = IdentityInterface(['i'])
    input = 'i'
    output = 'i'


class UnzipConverter(Converter):

    interface = UnzipDir()
    mem_gb = 12
    input = 'zipped'
    output = 'unzipped'


class ZipConverter(Converter):

    interface = ZipDir()
    mem_gb = 12
    input = 'dirname'
    output = 'zipped'


class TarGzConverter(Converter):

    interface = TarGzDir()
    mem_gb = 12
    input = 'dirname'
    output = 'zipped'


class UnTarGzConverter(Converter):

    interface = UnTarGzDir()
    mem_gb = 12
    input = 'gzipped'
    output = 'gunzipped'


# General formats
directory_format = FileFormat(name='directory', extension=None, directory=True)
text_format = FileFormat(name='text', extension='.txt')
json_format = FileFormat(name='json', extension='.json')

# Compressed formats
zip_format = FileFormat(name='zip', extension='.zip')
targz_format = FileFormat(name='targz', extension='.tar.gz')

standard_formats = [text_format, json_format, directory_format, zip_format,
                    targz_format]

# General image formats
gif_format = FileFormat(name='gif', extension='.gif')
png_format = FileFormat(name='png', extension='.png')
jpg_format = FileFormat(name='jpg', extension='.jpg')

# Set Converters
directory_format.set_converter(zip_format, UnzipConverter)
directory_format.set_converter(targz_format, UnTarGzConverter)
targz_format.set_converter(directory_format, TarGzConverter)
zip_format.set_converter(directory_format, ZipConverter)



import os
import tempfile
import os.path as op
from arcana.data import InputFilesetSpec, FilesetSpec, InputFilesets, Fileset
from arcana.data.file_format import (
    text_format, directory_format, zip_format)
from arcana.study.base import Study, StudyMetaClass
from arcana.utils.testing import BaseTestCase
from nipype.interfaces.utility import IdentityInterface
from arcana.utils.interfaces import ZipDir
from future.utils import with_metaclass
from unittest import TestCase


class TestConverterAvailability(TestCase):

    def test_find_converter(self):
        converter = zip_format.converter_from(directory_format)
        self.assertIsInstance(converter.interface, ZipDir)


class ConversionStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('text', text_format),
        InputFilesetSpec('directory', directory_format),
        InputFilesetSpec('zip', zip_format),
        FilesetSpec('text_from_text', text_format, 'conv_pipeline'),
        FilesetSpec('directory_from_zip_on_input', directory_format,
                    'conv_pipeline'),
        FilesetSpec('zip_from_directory_on_input', zip_format,
                    'conv_pipeline'),
        FilesetSpec('directory_from_zip_on_output', directory_format,
                    'conv_pipeline'),
        FilesetSpec('zip_from_directory_on_output', zip_format,
                    'conv_pipeline')]

    def conv_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='conv_pipeline',
            name_maps=name_maps,
            desc=("A pipeline that tests out various data format "
                         "conversions"))
        # No conversion from text to text format
        pipeline.add(
            'text_from_text',
            IdentityInterface(fields=['file']),
            inputs={
                'file': ('text', text_format)},
            outputs={
                'text_from_text': ('file', text_format)})
        # Convert from zip file to directory format on input
        pipeline.add(
            'directory_from_zip_on_input',
            IdentityInterface(fields=['file']),
            inputs={
                'file': ('zip', directory_format)},
            outputs={
                'directory_from_zip_on_input': ('file', directory_format)})
        # Convert from zip file to directory format on input
        pipeline.add(
            'directory_from_zip_on_output',
            IdentityInterface(fields=['file']),
            inputs={
                'file': ('zip', zip_format)},
            outputs={
                'directory_from_zip_on_output': ('file', zip_format)})
        # Convert from directory to zip format on input
        pipeline.add(
            'zip_from_directory_on_input',
            IdentityInterface(fields=['file']),
            inputs={
                'file': ('directory', zip_format)},
            outputs={
                'zip_from_directory_on_input': ('file', zip_format)})
        # Convert from directory to zip format on input
        pipeline.add(
            'zip_from_directory_on_output',
            IdentityInterface(fields=['file']),
            inputs={
                'file': ('directory', directory_format)},
            outputs={
                'zip_from_directory_on_output': ('file', directory_format)})
        return pipeline


class TestFormatConversions(BaseTestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        # Run BaseTestCase setUp
        super(TestFormatConversions, self).setUp()

    @property
    def INPUT_FILESETS(self):
        return {
            'text': 'text',
            'directory': self.input_directory,
            'zip': self.input_zip}

    @property
    def input_directory(self):
        path = op.join(self.tempdir, 'directory')
        if not op.exists(path):
            # Create directory
            os.makedirs(path)
            with open(op.join(path, 'dummy.txt'), 'w') as f:
                f.write('blah')
        return Fileset.from_path(path, format=directory_format)

    @property
    def input_zip(self):
        path = op.join(self.tempdir, 'zip.zip')
        if not op.exists(path):
            # Create zip file
            zipper = ZipDir()
            zipper.inputs.dirname = self.input_directory.path
            zipper.inputs.zipped = path
            zipper.run()
        return Fileset.from_path(path, format=zip_format)

    def test_format_conversions(self):
        study = self.create_study(
            ConversionStudy, 'conversion', [
                InputFilesets('text', 'text', text_format),
                InputFilesets('directory', 'directory', directory_format),
                InputFilesets('zip', 'zip', zip_format)])
        self.assertCreated(list(study.data('text_from_text'))[0])
        self.assertCreated(list(study.data('directory_from_zip_on_input'))[0])
        self.assertCreated(list(study.data('zip_from_directory_on_input'))[0])
        self.assertCreated(list(study.data('directory_from_zip_on_output'))[0])
        self.assertCreated(list(study.data('zip_from_directory_on_output'))[0])
